# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import re
import time
import json
import datetime
from lib389._constants import *
from lib389.properties import *
from lib389.cli_base import _get_arg
from lib389._entry import FormatDict
from lib389.utils import normalizeDN, ensure_bytes, ensure_str, ensure_dict_str, ensure_list_str
from lib389 import Entry, DirSrv, NoSuchEntryError, InvalidArgumentError
from lib389._mapped_object import DSLdapObject, DSLdapObjects


class Agreement(DSLdapObject):
    """A replication agreement from this server instance to
    another instance of directory server.

    - must attributes: [ 'cn' ]
    - RDN attribute: 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    csnpat = r'(.{8})(.{4})(.{4})(.{4})'
    csnre = re.compile(csnpat)

    def __init__(self, instance, dn=None, winsync=False):
        super(Agreement, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = [
            'cn',
        ]
        if winsync:
            self._create_objectclasses = [
                'top',
                'nsDSWindowsReplicationAgreement',
            ]
        else:
            self._create_objectclasses = [
                'top',
                'nsds5replicationagreement',
            ]
        self._protected = False

    def begin_reinit(self):
        """Begin a total reinit of the consumer. This will send
        our data to the server we are replicating too.
        """
        self.set('nsds5BeginReplicaRefresh', 'start')

    def check_reinit(self):
        """Check the status of a reinit. Returns done, inprogress, and error text. A correct
        reinit will return (True, False, False).

        :returns: tuple(done, error), where done, error are bool.
        """
        done = False
        error = False
        inprogress = False
        status = self.get_attr_val_utf8('nsds5ReplicaLastInitStatus')
        self._log.debug('agreement tot_init status: %s' % status)
        if not status:
            pass
        elif 'replica busy' in status:
            error = status
        elif 'Total update succeeded' in status:
            done = True
            inprogress = False
        elif 'Replication error' in status:
            error = status
        elif 'Total update in progress' in status:
            inprogress = True
        elif 'LDAP error' in status:
            error = status

        return (done, inprogress, error)

    def wait_reinit(self, timeout=300):
        """Wait for a reinit to complete. Returns done and error. A correct
        reinit will return (True, False).
        :param timeout: timeout value for how long to wait for the reinit
        :type timeout: int
        :returns: tuple(done, error), where done, error are bool.
        """
        done = False
        error = False
        count = 0
        while done is False and error is False:
            (done, inprogress, error) = self.check_reinit()
            if count > timeout and not done:
                error = True
            count = count + 2
            time.sleep(2)
        return (done, error)

    def get_name(self):
        return self.get_attr_val_utf8_l('cn')

    def get_agmt_maxcsn(self):
        """Get the agreement maxcsn from the database RUV entry
        :returns: CSN string if found, otherwise None is returned
        """
        from lib389.replica import Replicas
        suffix = self.get_attr_val_utf8(REPL_ROOT)
        agmt_name = self.get_attr_val_utf8('cn')
        replicas = Replicas(self._instance)
        replica = replicas.get(suffix)
        maxcsns = replica.get_ruv_agmt_maxcsns()

        if maxcsns is None or len(maxcsns) == 0:
            self._log.debug('get_agmt_maxcsn - Failed to get agmt maxcsn from RUV')
            return None

        for csn in maxcsns:
            comps = csn.split(';')
            if agmt_name == comps[1]:
                # same replica, get maxcsn
                if len(comps) < 6:
                    return None
                else:
                    return comps[5]

        self._log.debug('get_agmt_maxcsn - did not find matching agmt maxcsn from RUV')
        return None

    def get_consumer_maxcsn(self, binddn=None, bindpw=None):
        """Attempt to get the consumer's maxcsn from its database RUV entry
        :param binddn: Specifies a specific bind DN to use when contacting the remote consumer
        :type binddn: str
        :param bindpw: Password for the bind DN
        :type bindpw: str
        :returns: CSN string if found, otherwise "Unavailable" is returned
        """
        host = self.get_attr_val_utf8(AGMT_HOST)
        port = self.get_attr_val_utf8(AGMT_PORT)
        suffix = self.get_attr_val_utf8(REPL_ROOT)
        protocol = self.get_attr_val_utf8('nsds5replicatransportinfo').lower()

        result_msg = "Unavailable"

        # If we are using LDAPI we need to provide the credentials, otherwise
        # use the existing credentials
        if binddn is None:
            binddn = self._instance.binddn
        if bindpw is None:
            bindpw = self._instance.bindpw

        # Get the replica id from supplier to compare to the consumer's rid
        from lib389.replica import Replicas
        replicas = Replicas(self._instance)
        replica = replicas.get(suffix)
        rid = int(replica.get_attr_val_utf8(REPL_ID))

        # Open a connection to the consumer
        consumer = DirSrv(verbose=self._instance.verbose)
        args_instance[SER_HOST] = host
        if protocol == "ssl" or protocol == "ldaps":
            args_instance[SER_SECURE_PORT] = int(port)
        else:
            args_instance[SER_PORT] = int(port)
        args_instance[SER_ROOT_DN] = binddn
        args_instance[SER_ROOT_PW] = bindpw
        args_standalone = args_instance.copy()
        consumer.allocate(args_standalone)
        try:
            consumer.open()
        except ldap.INVALID_CREDENTIALS as e:
            raise(e)
        except ldap.LDAPError as e:
            self._log.debug('Connection to consumer ({}:{}) failed, error: {}'.format(host, port, e))
            return result_msg

        # Search for the tombstone RUV entry
        try:
            entry = consumer.search_s(suffix, ldap.SCOPE_SUBTREE,
                                      REPLICA_RUV_FILTER, ['nsds50ruv'])
            if not entry:
                self._log.debug("Failed to retrieve database RUV entry from consumer")
            else:
                elements = ensure_list_str(entry[0].getValues('nsds50ruv'))
                for ruv in elements:
                    if ('replica %d ' % rid) in ruv:
                        ruv_parts = ruv.split()
                        if len(ruv_parts) == 5:
                            result_msg = ruv_parts[4]
                        break
        except ldap.INVALID_CREDENTIALS as e:
            self._log.debug('Failed to search for the suffix ' +
                                     '({}) consumer ({}:{}) failed, error: {}'.format(
                                         suffix, host, port, e))
            raise(e)
        except ldap.LDAPError as e:
            self._log.debug('Failed to search for the suffix ' +
                                     '({}) consumer ({}:{}) failed, error: {}'.format(
                                         suffix, host, port, e))
        consumer.close()
        return result_msg

    def get_agmt_status(self, binddn=None, bindpw=None, return_json=False):
        """Return the status message
        :param binddn: Specifies a specific bind DN to use when contacting the remote consumer
        :type binddn: str
        :param bindpw: Password for the bind DN
        :type bindpw: str
        :returns: A status message about the replication agreement
        """
        con_maxcsn = "Unknown"
        try:
            agmt_maxcsn = self.get_agmt_maxcsn()
            agmt_status_json = self.get_attr_val_utf8_l(AGMT_UPDATE_STATUS_JSON)
            try:
                agmt_status = json.loads(agmt_status_json)
            except TypeError:
                # This may happen when dealing with an old version of directory server
                agmt_status_legacy = self.get_attr_val_utf8_l(AGMT_UPDATE_STATUS)
                agmt_status = { 'state': 'amber',
                                'message': f'Legacy replica message: is {agmt_status_legacy}' }
            except json.JSONDecodeError:
                agmt_status = { 'state': 'red', 
                                'message': f'Invalid status syntax: {agmt_status_json}' }
            if agmt_maxcsn is not None:
                try:
                    con_maxcsn = self.get_consumer_maxcsn(binddn=binddn, bindpw=bindpw)
                    if con_maxcsn:
                        if agmt_maxcsn == con_maxcsn:
                            if return_json:
                                return json.dumps({
                                    'msg': "In Synchronization",
                                    'agmt_maxcsn': agmt_maxcsn,
                                    'con_maxcsn': con_maxcsn,
                                    'state': agmt_status['state'],
                                    'reason': agmt_status['message']
                                }, indent=4)
                            else:
                                return "In Synchronization"
                except:
                    pass
            else:
                agmt_maxcsn = "Unknown"

            # Not in sync - attempt to discover the cause
            repl_msg = agmt_status['message']
            if self.get_attr_val_utf8_l(AGMT_UPDATE_IN_PROGRESS) == 'true':
                # Replication is on going - this is normal
                repl_msg = "Replication still in progress"
            elif "can't contact ldap" in agmt_status['message']:
                    # Consumer is down
                    repl_msg = "Consumer can not be contacted"

            if return_json:
                return json.dumps({
                    'msg': "Not in Synchronization",
                    'agmt_maxcsn': agmt_maxcsn,
                    'con_maxcsn': con_maxcsn,
                    'state': agmt_status['state'],
                    'reason': repl_msg
                }, indent=4)
            else:
                return ("Not in Synchronization: supplier " +
                        "(%s) consumer (%s) State (%s) Reason (%s)" %
                        (agmt_maxcsn, con_maxcsn, agmt_status['state'], repl_msg))
        except ldap.INVALID_CREDENTIALS as e:
            raise(e)
        except ldap.LDAPError as e:
            raise ValueError(str(e))

    def get_lag_time(self, suffix, agmt_name, binddn=None, bindpw=None):
        """Get the lag time between the supplier and the consumer
        :param suffix: The replication suffix
        :type suffix: str
        :param agmt_name: The name of the agreement
        :type agmt_name: str
        :param binddn: Specifies a specific bind DN to use when contacting the remote consumer
        :type binddn: str
        :param bindpw: Password for the bind DN
        :type bindpw: str
        :returns: A time-formated string of the the replication lag (HH:MM:SS).
        :raises: ValueError - if unable to get consumer's maxcsn
        """

        try:
            agmt_maxcsn = self.get_agmt_maxcsn()
            con_maxcsn = self.get_consumer_maxcsn(binddn=binddn, bindpw=bindpw)
        except ldap.LDAPError as e:
            raise ValueError("Unable to get lag time: " + str(e))

        if con_maxcsn is None:
            raise ValueError("Unable to get consumer's max csn")
        if con_maxcsn.lower() == "unavailable" or agmt_maxcsn is None:
            return "unavailable"

        # Extract the csn timestamps and compare them
        agmt_time = 0
        con_time = 0
        match = Agreement.csnre.match(agmt_maxcsn)
        if match:
            agmt_time = int(match.group(1), 16)
        match = Agreement.csnre.match(con_maxcsn)
        if match:
            con_time = int(match.group(1), 16)
        diff = con_time - agmt_time
        if diff < 0:
            lag = datetime.timedelta(seconds=-diff)
        else:
            lag = datetime.timedelta(seconds=diff)

        # Return a nice formated timestamp
        return "{:0>8}".format(str(lag))

    def status(self, winsync=False, just_status=False, use_json=False, binddn=None, bindpw=None, pwprompt=False):
        """Get the status of a replication agreement
        :param winsync: Specifies if the the agreement is a winsync replication agreement
        :type winsync: boolean
        :param just_status: Just return the status string and not all of the status attributes
        :type just_status: boolean
        :param use_json: Return the status in a JSON object
        :type use_json: boolean
        :param binddn: Specifies a specific bind DN to use when contacting the remote consumer
        :type binddn: str
        :param bindpw: Password for the bind DN
        :type bindpw: str
        :param pwprompt: If binddn or bindpw is None, ask for them interactively
        :type pwprompt: boolean
        :returns: A status message
        :raises: ValueError - if failing to get agmt status
        """
        status_attrs_dict = self.get_all_attrs()
        status_attrs_dict = dict((k.lower(), v) for k, v in list(status_attrs_dict.items()))

        # We need a bind DN and passwd so we can query the consumer.  If this is an LDAPI
        # connection, and the consumer does not allow anonymous access to the tombstone
        # RUV entry under the suffix, then we can't get the status.  So in this case we
        # need to provide a DN and password.
        if not winsync:
            if pwprompt:
                host = self.get_attr_val_utf8(AGMT_HOST)
                port = self.get_attr_val_utf8(AGMT_PORT)
                suffix = self.get_attr_val_utf8(REPL_ROOT)
                if binddn is None:
                    binddn = _get_arg(None, msg=f"Enter bind DN for the replicated suffix ({suffix}) on {host}:{port}")
                if bindpw is None:
                    bindpw = _get_arg(None, msg=f"Enter password for ({binddn}) to the replicated suffix ({suffix}) on {host}:{port}", hidden=True)
            try:
                status = self.get_agmt_status(binddn=binddn, bindpw=bindpw)
            except ldap.INVALID_CREDENTIALS as e:
                raise(e)
            except ValueError as e:
                status = str(e)
            if just_status:
                if use_json:
                    return (json.dumps(status, indent=4))
                else:
                    return status

            # Get the lag time
            suffix = ensure_str(status_attrs_dict['nsds5replicaroot'][0])
            agmt_name = ensure_str(status_attrs_dict['cn'][0])
            lag_time = self.get_lag_time(suffix, agmt_name, binddn=binddn, bindpw=bindpw)
        else:
            lag_time = "Not available for Winsync agreements"
            status = "Not available for Winsync agreements"

        # handle the attributes that are not always set in the agreement
        if 'nsds5replicaenabled' not in status_attrs_dict:
            status_attrs_dict['nsds5replicaenabled'] = ['on']
        if 'nsds5agmtmaxcsn' not in status_attrs_dict:
            status_attrs_dict['nsds5agmtmaxcsn'] = ["unavailable"]
        if 'nsds5replicachangesskippedsince' not in status_attrs_dict:
            status_attrs_dict['nsds5replicachangesskippedsince'] = ["unavailable"]
        if 'nsds5beginreplicarefresh' not in status_attrs_dict:
            status_attrs_dict['nsds5beginreplicarefresh'] = [""]
        if 'nsds5replicalastinitstatus' not in status_attrs_dict:
            status_attrs_dict['nsds5replicalastinitstatus'] = ["unavailable"]
        if 'nsds5replicachangessentsincestartup' not in status_attrs_dict:
            status_attrs_dict['nsds5replicachangessentsincestartup'] = ['0']
        if ensure_str(status_attrs_dict['nsds5replicachangessentsincestartup'][0]) == '':
            status_attrs_dict['nsds5replicachangessentsincestartup'] = ['0']

        consumer = "{}:{}".format(ensure_str(status_attrs_dict['nsds5replicahost'][0]),
                                  ensure_str(status_attrs_dict['nsds5replicaport'][0]))

        # Case sensitive?
        if use_json:
            result = {
                      'agmt-name': ensure_list_str(status_attrs_dict['cn']),
                      'replica': [consumer],
                      'replica-enabled': ensure_list_str(status_attrs_dict['nsds5replicaenabled']),
                      'update-in-progress': ensure_list_str(status_attrs_dict['nsds5replicaupdateinprogress']),
                      'last-update-start': ensure_list_str(status_attrs_dict['nsds5replicalastupdatestart']),
                      'last-update-end': ensure_list_str(status_attrs_dict['nsds5replicalastupdateend']),
                      'number-changes-sent': ensure_list_str(status_attrs_dict['nsds5replicachangessentsincestartup']),
                      'number-changes-skipped': ensure_list_str(status_attrs_dict['nsds5replicachangesskippedsince']),
                      'last-update-status': ensure_list_str(status_attrs_dict['nsds5replicalastupdatestatus']),
                      'last-init-start': ensure_list_str(status_attrs_dict['nsds5replicalastinitstart']),
                      'last-init-end': ensure_list_str(status_attrs_dict['nsds5replicalastinitend']),
                      'last-init-status': ensure_list_str(status_attrs_dict['nsds5replicalastinitstatus']),
                      'reap-active': ensure_list_str(status_attrs_dict['nsds5replicareapactive']),
                      'replication-status': [status],
                      'replication-lag-time': [lag_time]
                }
            return (json.dumps(result, indent=4))
        else:
            retstr = (
                "Status For Agreement: \"%(cn)s\" (%(nsDS5ReplicaHost)s:"
                "%(nsDS5ReplicaPort)s)" "\n"
                "Replica Enabled: %(nsds5ReplicaEnabled)s" "\n"
                "Update In Progress: %(nsds5replicaUpdateInProgress)s" "\n"
                "Last Update Start: %(nsds5replicaLastUpdateStart)s" "\n"
                "Last Update End: %(nsds5replicaLastUpdateEnd)s" "\n"
                "Number Of Changes Sent: %(nsds5replicaChangesSentSinceStartup)s"
                "\n"
                "Number Of Changes Skipped: %(nsds5replicaChangesSkippedSince"
                "Startup)s" "\n"
                "Last Update Status: %(nsds5replicaLastUpdateStatus)s" "\n"
                "Last Init Start: %(nsds5ReplicaLastInitStart)s" "\n"
                "Last Init End: %(nsds5ReplicaLastInitEnd)s" "\n"
                "Last Init Status: %(nsds5ReplicaLastInitStatus)s" "\n"
                "Reap Active: %(nsds5ReplicaReapActive)s" "\n"
            )
            # FormatDict manages missing fields in string formatting
            entry_data = ensure_dict_str(status_attrs_dict)
            result = retstr % FormatDict(entry_data)
            result += "Replication Status: {}\n".format(status)
            result += "Replication Lag Time: {}\n".format(lag_time)
            return result

    def pause(self):
        """Pause outgoing changes from this server to consumer. Note
        that this does not pause the consumer, only that changes will
        not be sent from this supplier to consumer: the consumer may still
        receive changes from other replication paths!
        """
        self.set('nsds5ReplicaEnabled', 'off')

    def resume(self):
        """Resume sending updates from this supplier to consumer directly.
        """
        self.set('nsds5ReplicaEnabled', 'on')

    def set_wait_for_async_results(self, value):
        """Set nsDS5ReplicaWaitForAsyncResults to value.

        :param value: Time in milliseconds.
        :type value: str
        """
        self.replace('nsDS5ReplicaWaitForAsyncResults', value)

    def remove_wait_for_async_results(self):
        """Reset nsDS5ReplicaWaitForAsyncResults to default.
        """
        self.remove_all('nsDS5ReplicaWaitForAsyncResults')

    def get_wait_for_async_results_utf8(self):
        """Get the current value of nsDS5ReplicaWaitForAsyncResults.

        :returns: str
        """
        return self.get_attr_val_utf8('nsDS5ReplicaWaitForAsyncResults')

    def set_flowcontrolwindow(self, value):
        """Set nsds5ReplicaFlowControlWindow to value.

        :param value: During total update Number of entries to send without waiting ack
        :type value: str
        """
        self.replace('nsds5ReplicaFlowControlWindow', value)

class WinsyncAgreement(Agreement):
    """A replication agreement from this server instance to
    another instance of directory server.

    - must attributes: [ 'cn' ]
    - RDN attribute: 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(Agreement, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = [
            'cn',
        ]
        self._create_objectclasses = [
                'top',
                'nsDSWindowsReplicationAgreement',
            ]

        self._protected = False


class Agreements(DSLdapObjects):
    """Represents the set of agreements configured on this instance.
    There are two possible ways to use this interface.

    The first is as the set of agreements on the server for all
    replicated suffixes. IE:

    agmts = Agreements(inst).

    However, this will NOT allow new agreements to be created, as
    agreements must be related to a replica.

    The second is Agreements related to a replica. For this method
    you must use:

    replica = Replicas(inst).get(<suffix>)
    agmts = Agreements(inst, replica.dn)


    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: The base dn to search.
    :type basedn: str
    :param rdn: The rdn relative to cn=mapping tree to search.
    :type rdn: str
    """

    def __init__(self, instance, basedn=DN_MAPPING_TREE, rdn=None, winsync=False):
        super(Agreements, self).__init__(instance)
        if winsync:
            self._childobject = WinsyncAgreement
            self._objectclasses = ['nsDSWindowsReplicationAgreement']
        else:
            self._childobject = Agreement
            self._objectclasses = ['nsds5replicationagreement']
        self._filterattrs = ['cn', 'nsDS5ReplicaRoot']
        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = "%s,%s" % (rdn, basedn)

    def _validate(self, rdn, properties):
        """ An internal implementation detail of create verification. You should
        never call this directly.
        """
        if self._basedn == DN_MAPPING_TREE:
            raise ldap.UNWILLING_TO_PERFORM("Refusing to create agreement in %s" % DN_MAPPING_TREE)
        return super(Agreements, self)._validate(rdn, properties)


class AgreementLegacy(object):
    """An object that helps to work with agreement entry

    :param conn: An instance
    :type conn: lib389.DirSrv
    """

    ALWAYS = '0000-2359 0123456'
    NEVER = '2358-2359 0'

    proxied_methods = 'search_s getEntry'.split()

    def __init__(self, conn):
        self.conn = conn
        self.log = conn.log

    def __getattr__(self, name):
        if name in Agreement.proxied_methods:
            return DirSrv.__getattr__(self.conn, name)

    def status(self, agreement_dn, just_status=False):
        """Get a formatted string with the replica status

        :param agreement_dn: DN of the replica agreement
        :type agreement_dn: str
        :param just_status: If True, returns just status
        :type just_status: bool
        :returns: str -- See below
        :raises: NoSuchEntryError - if agreement_dn is an unknown entry

        :example:
            ::

                Status for meTo_localhost.localdomain:50389 agmt
                    localhost.localdomain:50389
                Update in progress: TRUE
                Last Update Start: 20131121132756Z
                Last Update End: 0
                Num. Changes Sent: 1:10/0
                Num. changes Skipped: None
                Last update Status: Error (0) Replica acquired successfully:
                    Incremental update started
                Init in progress: None
                Last Init Start: 0
                Last Init End: 0
                Last Init Status: None
                Reap Active: 0
        """

        attrlist = ['cn', 'nsds5BeginReplicaRefresh', 'nsds5ReplicaRoot',
                    'nsds5replicaUpdateInProgress',
                    'nsds5ReplicaLastInitStatus', 'nsds5ReplicaLastInitStart',
                    'nsds5ReplicaLastInitEnd', 'nsds5replicaReapActive',
                    'nsds5replicaLastUpdateStart', 'nsds5replicaLastUpdateEnd',
                    'nsds5replicaChangesSentSinceStartup',
                    'nsds5replicaLastUpdateStatus',
                    'nsds5replicaChangesSkippedSinceStartup',
                    'nsds5ReplicaHost', 'nsds5ReplicaPort',
                    'nsds5ReplicaEnabled', 'nsds5AgmtMaxCSN']
        try:
            ent = self.conn.getEntry(
                agreement_dn, ldap.SCOPE_BASE, "(objectclass=*)", attrlist)
        except NoSuchEntryError:
            raise NoSuchEntryError(
                "Error reading status from agreement", agreement_dn)
        else:
            status = self.conn.getReplAgmtStatus(ent)
            if just_status:
                return status

            retstr = (
                "Status for %(cn)s agmt %(nsDS5ReplicaHost)s:"
                "%(nsDS5ReplicaPort)s" "\n"
                "Replica Enabled: %(nsds5ReplicaEnabled)s" "\n"
                "Agreement maxCSN: %(nsds5AgmtMaxCSN)s" "\n"
                "Update in progress: %(nsds5replicaUpdateInProgress)s" "\n"
                "Last Update Start: %(nsds5replicaLastUpdateStart)s" "\n"
                "Last Update End: %(nsds5replicaLastUpdateEnd)s" "\n"
                "Num. Changes Sent: %(nsds5replicaChangesSentSinceStartup)s"
                "\n"
                "Num. changes Skipped: %(nsds5replicaChangesSkippedSince"
                "Startup)s" "\n"
                "Last update Status: %(nsds5replicaLastUpdateStatus)s" "\n"
                "Init in progress: %(nsds5BeginReplicaRefresh)s" "\n"
                "Last Init Start: %(nsds5ReplicaLastInitStart)s" "\n"
                "Last Init End: %(nsds5ReplicaLastInitEnd)s" "\n"
                "Last Init Status: %(nsds5ReplicaLastInitStatus)s" "\n"
                "Reap Active: %(nsds5ReplicaReapActive)s" "\n"
            )
            # FormatDict manages missing fields in string formatting
            entry_data = ensure_dict_str(ent.data)
            result = retstr % FormatDict(entry_data)
            result += "Replication Status: %s\n" % status
            return result

    def _check_interval(self, interval):
        """Check the interval for schedule replication is valid:
        HH [0..23]
        MM [0..59]
        DAYS [0-6]{1,7}

        :param interval: - 'HHMM-HHMM D+' With D=[0123456]+
                          - Agreement.ALWAYS
                          - Agreement.NEVER
        :type interval: str

        :returns: None
        :raises: ValueError - if the interval is illegal
        """

        c = re.compile(re.compile('^([0-9][0-9])([0-9][0-9])-([0-9][0-9])' +
                                  '([0-9][0-9]) ([0-6]{1,7})$'))
        if not c.match(interval):
            raise ValueError("Bad schedule format %r" % interval)
        schedule = c.split(interval, c.groups)

        # check the hours
        hour = int(schedule[1])
        if ((hour < 0) or (hour > 23)):
            raise ValueError("Bad schedule format %r: illegal hour %d" %
                             (interval, hour))
        hour = int(schedule[3])
        if ((hour < 0) or (hour > 23)):
            raise ValueError("Bad schedule format %r: illegal hour %d" %
                             (interval, hour))
        if int(schedule[1]) > int(schedule[3]):
            raise ValueError("Bad schedule (start HOUR larger than end HOUR)" +
                             " %r: illegal hour %d" %
                             (interval, int(schedule[1])))

        # check the minutes
        minute = int(schedule[2])
        if ((minute < 0) or (minute > 59)):
            raise ValueError("Bad schedule format %r: illegal minute %d" %
                             (interval, minute))
        minute = int(schedule[4])
        if ((minute < 0) or (minute > 59)):
            raise ValueError("Bad schedule format %r: illegal minute %d" %
                             (interval, minute))

    def schedule(self, agmtdn=None, interval=ALWAYS):
        """Schedule the replication agreement

        :param agmtdn: DN of the replica agreement
        :type agmtdn: str
        :param interval: - 'HHMM-HHMM D+' With D=[0123456]+
                          - Agreement.ALWAYS
                          - Agreement.NEVER
        :type interval: str

        :returns: None
        :raises: - ValueError - if interval is not valid;
                  - ldap.NO_SUCH_OBJECT - if agmtdn does not exist
        """
        if not agmtdn:
            raise InvalidArgumentError("agreement DN is missing")

        # check the validity of the interval
        if interval != Agreement.ALWAYS and interval != Agreement.NEVER:
            self._check_interval(interval)

        # Check if the replica agreement exists
        try:
            self.conn.getEntry(agmtdn, ldap.SCOPE_BASE)
        except ldap.NO_SUCH_OBJECT:
            raise

        # update it
        self.log.info("Schedule replication agreement %s", agmtdn)
        mod = [(ldap.MOD_REPLACE, 'nsds5replicaupdateschedule', [ensure_bytes(interval)])]
        self.conn.modify_s(agmtdn, mod)

    def getProperties(self, agmnt_dn=None, properties=None):
        """Get a dictionary of the requested properties.
        If properties parameter is missing, it returns all the properties.

        :param agmtdn: DN of the replica agreement
        :type agmtdn: str
        :param properties: List of properties name
        :type properties: list

        :returns: Returns a dictionary of the properties
        :raises: - ValueError - if invalid property name
                  - ldap.NO_SUCH_OBJECT - if agmtdn does not exist
                  - InvalidArgumentError - missing mandatory argument

        :supported properties are:
                RA_NAME, RA_SUFFIX, RA_BINDDN, RA_BINDPW, RA_METHOD,
                RA_DESCRIPTION, RA_SCHEDULE, RA_TRANSPORT_PROT, RA_FRAC_EXCLUDE,
                RA_FRAC_EXCLUDE_TOTAL_UPDATE, RA_FRAC_STRIP, RA_CONSUMER_PORT,
                RA_CONSUMER_HOST, RA_CONSUMER_TOTAL_INIT, RA_TIMEOUT, RA_CHANGES
        """

        if not agmnt_dn:
            raise InvalidArgumentError("agmtdn is a mandatory argument")

        #
        # prepare the attribute list to retrieve from the RA
        # if properties is None, all RA attributes are retrieved
        #
        attrs = []
        if properties:
            for prop_name in properties:
                prop_attr = RA_PROPNAME_TO_ATTRNAME[prop_name]
                if not prop_attr:
                    raise ValueError("Improper property name: %s ", prop_name)
                attrs.append(prop_attr)

        filt = "(objectclass=*)"
        result = {}
        try:
            entry = self.conn.getEntry(agmnt_dn, ldap.SCOPE_BASE, filt, attrs)

            # Build the result from the returned attributes
            for attr in entry.getAttrs():
                # given an attribute name retrieve the property name
                props = [k for k, v in RA_PROPNAME_TO_ATTRNAME.items()
                         if v.lower() == attr.lower()]

                # If this attribute is present in the RA properties, adds it to
                # result
                if len(props) > 0:
                    result[props[0]] = entry.getValues(attr)

        except ldap.NO_SUCH_OBJECT:
            raise

        return result

    def setProperties(self, suffix=None, agmnt_dn=None, agmnt_entry=None,
                      properties=None):
        """Set the properties of the agreement entry. If an 'agmnt_entry'
        is provided, it updates the entry, else it updates the entry on
        the server. If the 'agmnt_dn' is provided it retrieves the entry
        using it, else it retrieve the agreement using the 'suffix'.

        :param suffix: Suffix stored in that agreement (online update)
        :type suffix: str
        :param agmnt_dn: DN of the agreement (online update)
        :type agmnt_dn: str
        :param agmnt_entry: Entry of a agreement (offline update)
        :type agmnt_entry: lib389.Entry
        :param properties: Dictionary of properties
        :type properties: dict

        :returns: None
        :raises: - ValueError - if invalid properties
                  - ValueError - if invalid agreement_entry
                  - ValueError - if agmnt_dn or suffix are not associated to a replica
                  - InvalidArgumentError - missing mandatory argument

        :supported properties are:
                RA_NAME, RA_SUFFIX, RA_BINDDN, RA_BINDPW, RA_METHOD,
                RA_DESCRIPTION, RA_SCHEDULE, RA_TRANSPORT_PROT, RA_FRAC_EXCLUDE,
                RA_FRAC_EXCLUDE_TOTAL_UPDATE, RA_FRAC_STRIP, RA_CONSUMER_PORT,
                RA_CONSUMER_HOST, RA_CONSUMER_TOTAL_INIT, RA_TIMEOUT, RA_CHANGES
        """

        # No properties provided
        if len(properties) == 0:
            return

        # check that the given properties are valid
        for prop in properties:
            # skip the prefix to add/del value
            if not inProperties(prop, RA_PROPNAME_TO_ATTRNAME):
                raise ValueError("unknown property: %s" % prop)
            else:
                self.log.debug("setProperties: %s:%s",
                               prop, properties[prop])

        # At least we need to have suffix/agmnt_dn/agmnt_entry
        if not suffix and not agmnt_dn and not agmnt_entry:
            raise InvalidArgumentError("suffix and agmnt_dn and agmnt_entry "
                                       "are missing")

        # TODO
        if suffix:
            raise NotImplementedError

        # The caller provides a set of properties to set into a replica entry
        if agmnt_entry:
            if not isinstance(agmnt_entry, Entry):
                raise ValueError("invalid instance of the agmnt_entry")

            # that is fine, now set the values
            for prop in properties:
                val = rawProperty(prop)

                # for Entry update it is a replace
                agmnt_entry.update({RA_PROPNAME_TO_ATTRNAME[val]:
                                   properties[prop]})

            return

        # for each provided property build the mod
        mod = []
        for prop in properties:

            # retrieve/check the property name
            # and if the value needs to be added/deleted/replaced
            if prop.startswith('+'):
                mod_type = ldap.MOD_ADD
                prop_name = prop[1:]
            elif prop.startswith('-'):
                mod_type = ldap.MOD_DELETE
                prop_name = prop[1:]
            else:
                mod_type = ldap.MOD_REPLACE
                prop_name = prop

            attr = RA_PROPNAME_TO_ATTRNAME[prop_name]
            if not attr:
                raise ValueError("invalid property name %s" % prop_name)

            # Now do the value checking we can do
            if prop_name == RA_SCHEDULE:
                self._check_interval(properties[prop])

            mod.append((mod_type, attr, ensure_bytes(properties[prop])))

        # Now time to run the effective modify
        self.conn.modify_s(agmnt_dn, mod)

    def list(self, suffix=None, consumer_host=None, consumer_port=None,
             agmtdn=None):
        """Returns the search result of the replica agreement(s) under the
        replica (replicaRoot is 'suffix').

        Either 'suffix' or 'agmtdn' need to be specfied.
        'consumer_host' and 'consumer_port' are either not specified or
        specified both.

        If 'agmtdn' is specified, it returns the search result entry of
        that replication agreement, else if consumer host/port are specified
        it returns the replica agreements toward that consumer host:port.

        Finally if neither 'agmtdn' nor 'consumser host/port' are
        specifies it returns all the replica agreements under the replica
        (replicaRoot is 'suffix').

        :param suffix: The suffix targeted by the total update
        :type suffix: str
        :param consumer_host: Hostname of the consumer
        :type consumer_host: str
        :param consumer_port: Port of the consumer
        :type consumer_port: int
        :param agmtdn: DN of the replica agreement
        :type agmtdn: str

        :returns: Search result of the replica agreements
        :raises: - InvalidArgument - if missing mandatory argument
                    (agmtdn or suffix, then host and port)
                  - ValueError - if some properties are not valid
                  - NoSuchEntryError - If no replica defined for the suffix
        """

        if not suffix and not agmtdn:
            raise InvalidArgumentError("suffix or agmtdn are required")

        if (consumer_host and not consumer_port) or (not consumer_host and
                                                     consumer_port):
            raise InvalidArgumentError(
                "consumer host/port are required together")

        if agmtdn:
            # easy case, just return the RA
            filt = "objectclass=*"
            return self.conn.search_s(agmtdn, ldap.SCOPE_BASE, filt)
        else:
            # Retrieve the replica
            replica_entries = self.conn.replica.list(suffix)
            if not replica_entries:
                raise NoSuchEntryError("Error: no replica set up for suffix "
                                       "(%s)" % suffix)
            replica_entry = replica_entries[0]

            # Now returns the replica agreement for that suffix that replicates
            # to consumer host/port
            if consumer_host and consumer_port:
                '''
                Currently python does not like long continuous lines when it
                comes to string formatting, so we need to separate each line
                like this.
                '''
                filt = "(&(|(objectclass=%s)" % RA_OBJECTCLASS_VALUE
                filt += "(objectclass=%s))" % RA_WINDOWS_OBJECTCLASS_VALUE
                filt += "(%s=%s)" % (RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_HOST],
                                     consumer_host)
                filt += "(%s=%d))" % (
                    RA_PROPNAME_TO_ATTRNAME[RA_CONSUMER_PORT], consumer_port)
            else:
                filt = ("(|(objectclass=%s)(objectclass=%s))" %
                        (RA_OBJECTCLASS_VALUE, RA_WINDOWS_OBJECTCLASS_VALUE))

            return self.conn.search_s(replica_entry.dn, ldap.SCOPE_ONELEVEL,
                                      filt)

    def create(self, suffix=None, host=None, port=None, properties=None,
               winsync=False):
        """Create (and return) a replication agreement from self to consumer.
        Self is the supplier.

        :param suffix: Replication Root
        :type suffix: str
        :param host: Consumer host
        :type host: str
        :param port: Consumer port
        :type port: int
        :param winsync: Identifies the agree as a WinSync agreement
        :type winsync: bool
        :param properties: Agreement properties
        :type properties: dict

        :returns: DN of the created agreement
        :raises: - InvalidArgumentError - If the suffix is missing
                 - NoSuchEntryError - if a replica doesn't exist for that suffix
                 - ldap.LDAPError - ldap error
        """

        # Check we have a suffix [ mandatory ]
        if not suffix:
            self.log.warning("create: suffix is missing")
            raise InvalidArgumentError('suffix is mandatory')

        if not properties:
            properties = {}

        # Compute the normalized suffix to be set in RA entry
        properties[RA_SUFFIX] = normalizeDN(suffix)

        # Adding agreement under the replica entry
        replica_entries = self.conn.replica.list(suffix)
        if not replica_entries:
            raise NoSuchEntryError(
                "Error: no replica set up for suffix: %s" % suffix)
        replica = replica_entries[0]

        # Define agreement entry
        if RA_NAME not in properties:
            properties[RA_NAME] = 'meTo_%s:%s' % (host, port)
        dn_agreement = ','.join(["cn=%s" % properties[RA_NAME], replica.dn])

        # Set the required properties(if not already set)
        if RA_BINDDN not in properties:
            properties[RA_BINDDN] = defaultProperties[REPLICATION_BIND_DN]
        if RA_BINDPW not in properties:
            properties[RA_BINDPW] = defaultProperties[REPLICATION_BIND_PW]
        if RA_METHOD not in properties:
            properties[RA_METHOD] = defaultProperties[REPLICATION_BIND_METHOD]
        if RA_TRANSPORT_PROT not in properties:
            properties[RA_TRANSPORT_PROT] = \
                defaultProperties[REPLICATION_TRANSPORT]
        if RA_TIMEOUT not in properties:
            properties[RA_TIMEOUT] = defaultProperties[REPLICATION_TIMEOUT]
        if RA_DESCRIPTION not in properties:
            properties[RA_DESCRIPTION] = properties[RA_NAME]
        if RA_CONSUMER_HOST not in properties:
            properties[RA_CONSUMER_HOST] = host
        if RA_CONSUMER_PORT not in properties:
            properties[RA_CONSUMER_PORT] = str(port)

        # This is probably unnecessary because
        # we can just raise ALREADY_EXISTS
        try:
            entry = self.conn.getEntry(dn_agreement, ldap.SCOPE_BASE)
            self.log.warning("Agreement already exists: %r", dn_agreement)
            return dn_agreement
        except ldap.NO_SUCH_OBJECT:
            entry = None

        # Iterate over the properties, adding them to the entry
        entry = Entry(dn_agreement)
        entry.update({'objectclass': ["top", RA_OBJECTCLASS_VALUE]})
        for prop in properties:
            entry.update({RA_PROPNAME_TO_ATTRNAME[prop]: properties[prop]})

        # we make a copy here because we cannot change
        # the passed in properties dict
        propertiescopy = {}
        if properties:
            import copy
            propertiescopy = copy.deepcopy(properties)

        # Check if this a Winsync Agreement
        if winsync:
            self.conn.setupWinSyncAgmt(propertiescopy, entry)

        try:
            self.log.debug("Adding replica agreement: [%r]", entry)
            self.conn.add_s(entry)
        except ldap.LDAPError as e:
            self.log.fatal('Failed to add replication agreement: %s', e)
            raise e

        entry = self.conn.waitForEntry(dn_agreement)
        if entry:
            # More verbose but shows what's going on
            if 'chain' in propertiescopy:
                raise NotImplementedError
                '''
                chain_args = {
                    'suffix': suffix,
                    'binddn': binddn,
                    'bindpw': bindpw
                }
                # Work on `self` aka producer
                if replica.nsds5replicatype == SUPPLIER_TYPE:
                    self.conn.setupChainingFarm(**chain_args)
                # Work on `consumer`
                # TODO - is it really required?
                if replica.nsds5replicatype == LEAF_TYPE:
                    chain_args.update({
                        'isIntermediate': 0,
                        'urls': self.conn.toLDAPURL(),
                        'args': propertiescopy['chainargs']
                    })
                    consumer.setupConsumerChainOnUpdate(**chain_args)
                elif replica.nsds5replicatype == HUB_TYPE:
                    chain_args.update({
                        'isIntermediate': 1,
                        'urls': self.conn.toLDAPURL(),
                        'args': propertiescopy['chainargs']
                    })
                    consumer.setupConsumerChainOnUpdate(**chain_args)
                '''
        return dn_agreement

    def delete(self, suffix=None, consumer_host=None, consumer_port=None,
               agmtdn=None):
        """Delete a replication agreement

        :param suffix: The suffix that the agreement is configured for
        :type suffix: str
        :param consumer_host: Host of the server that the agreement points to
        :type consumer_host: str
        :param consumer_port: Port of the server that the agreement points to
        :type consumer_port: int
        :param agmtdn: DN of the replica agreement
        :type agmtdn: str

        :returns: None
        :raises: - ldap.LDAPError - for ldap operation failures
                  - TypeError - if too many agreements were found
                  - NoSuchEntryError - if no agreements were found
        """

        if not (suffix and consumer_host and consumer_port) and not agmtdn:
            raise InvalidArgumentError("Suffix with consumer_host and consumer_port" +
                                       " or agmtdn are required")

        agmts = self.list(suffix, consumer_host, consumer_port, agmtdn)

        if agmts:
            if len(agmts) > 1:
                raise TypeError('Too many agreements were found')
            else:
                # Delete the agreement
                try:
                    agmt_dn = agmts[0].dn
                    self.conn.delete_s(agmt_dn)
                    self.log.info('Agreement (%s) was successfully removed', agmt_dn)
                except ldap.LDAPError as e:
                    self.log.error('Failed to delete agreement (%s), '
                                   'error: %s', agmt_dn, e)
                    raise
        else:
            raise NoSuchEntryError('No agreements were found')

    def init(self, suffix=None, consumer_host=None, consumer_port=None):
        """Trigger a total update of the consumer replica
        - self is the supplier,
        - consumer is a DirSrv object (consumer can be a supplier)
        - cn_format - use this string to format the agreement name

        :param suffix: The suffix targeted by the total update [mandatory]
        :type suffix: str
        :param consumer_host: Hostname of the consumer [mandatory]
        :type consumer_host: str
        :param consumer_port: Port of the consumer [mandatory]
        :type consumer_port: int

        :returns: None
        :raises: InvalidArgument - if missing mandatory argument
        """

        #
        # check the required parameters are set
        #
        if not suffix:
            self.log.fatal("initAgreement: suffix is missing")
            raise InvalidArgumentError('suffix is mandatory argument')

        nsuffix = normalizeDN(suffix)

        if not consumer_host:
            self.log.fatal("initAgreement: host is missing")
            raise InvalidArgumentError('host is mandatory argument')

        if not consumer_port:
            self.log.fatal("initAgreement: port is missing")
            raise InvalidArgumentError('port is mandatory argument')
        #
        # check the replica agreement already exist
        #
        replica_entries = self.conn.replica.list(suffix)
        if not replica_entries:
            raise NoSuchEntryError(
                "Error: no replica set up for suffix " + suffix)
        replica_entry = replica_entries[0]
        self.log.debug("initAgreement: looking for replica agreements "
                       "under %s", replica_entry.dn)
        try:
            '''
            Currently python does not like long continuous lines when it
            comes to string formatting, so we need to separate each line
            like this.
            '''
            filt = "(&(objectclass=nsds5replicationagreement)"
            filt += "(nsds5replicahost=%s)" % consumer_host
            filt += "(nsds5replicaport=%d)" % consumer_port
            filt += "(nsds5replicaroot=%s))" % nsuffix

            entry = self.conn.getEntry(replica_entry.dn, ldap.SCOPE_ONELEVEL,
                                       filt)
        except ldap.NO_SUCH_OBJECT:
            msg = ('initAgreement: No replica agreement to ' +
                   '{host}:{port} for suffix {suffix}'.format(
                       host=consumer_host, port=consumer_port, suffix=nsuffix))
            self.log.fatal(msg)
            raise

        #
        # trigger the total init
        #
        self.log.info("Starting total init %s", entry.dn)
        mod = [(ldap.MOD_ADD, 'nsds5BeginReplicaRefresh', ensure_bytes('start'))]
        self.conn.modify_s(entry.dn, mod)

    def pause(self, agmtdn, interval=NEVER):
        """Pause this replication agreement.  This replication agreement
        will send no more changes.  Use the resume() method to "unpause".

        It tries to disable the replica agreement. If it fails (not
        implemented in all version),

        It uses the schedule() with interval '2358-2359 0'

        :param agmtdn: agreement dn
        :type agmtdn: str
        :param interval: - 'HHMM-HHMM D+' With D=[0123456]+
                          - Agreement.ALWAYS
                          - Agreement.NEVER
                          - Default is NEVER
        :type interval: str

        :returns: None
        :raises: ValueError - if interval is not valid
        """

        self.log.info("Pausing replication %s", agmtdn)
        mod = [(ldap.MOD_REPLACE, 'nsds5ReplicaEnabled', [b'off'])]
        self.conn.modify_s(ensure_str(agmtdn), mod)

        # Allow a little time for repl agmt thread to stop
        time.sleep(5)

    def resume(self, agmtdn, interval=ALWAYS):
        """Resume a paused replication agreement, paused with the "pause" method.
        It tries to enabled the replica agreement. If it fails
        (not implemented in all versions)
        It uses the schedule() with interval '0000-2359 0123456'.

        :param agmtdn: agreement dn
        :type agmtdn: str
        :param interval: - 'HHMM-HHMM D+' With D=[0123456]+
                          - Agreement.ALWAYS
                          - Agreement.NEVER
                          - Default is NEVER
        :type interval: str

        :returns: None
        :raises: - ValueError - if interval is not valid
                  - ldap.NO_SUCH_OBJECT - if agmtdn does not exist
        """

        self.log.info("Resuming replication %s", agmtdn)
        mod = [(ldap.MOD_REPLACE, 'nsds5ReplicaEnabled', [b'on'])]
        self.conn.modify_s(ensure_str(agmtdn), mod)

        # Allow a little time for the repl agmt thread to start
        time.sleep(2)

    def changes(self, agmnt_dn):
        """Get a number of changes sent by this agreement.

        :param agmtdn: agreement dn
        :type agmtdn: str

        :returns: Number of changes
        :raises: NoSuchEntryError - if agreement entry with changes
                  attribute is not found
        """

        retval = 0
        try:
            ent = self.conn.getEntry(ensure_str(agmnt_dn), ldap.SCOPE_BASE,
                                     "(objectclass=*)",
                                     [RA_PROPNAME_TO_ATTRNAME[RA_CHANGES]])
        except:
            raise NoSuchEntryError(
                "Error reading status from agreement", agmnt_dn)

        if ent.nsds5replicaChangesSentSinceStartup:
            val = ent.nsds5replicaChangesSentSinceStartup
            items = val.split(ensure_bytes(' '))
            if len(items) == 1:
                retval = int(items[0])
            else:
                for item in items:
                    ary = item.split(ensure_bytes(":"))
                    if ary and len(ary) > 1:
                        retval = retval + int(ary[1].split(ensure_bytes("/"))[0])
        return retval



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.backend import Backends
from lib389.utils import ensure_str
from lib389.pwpolicy import PwPolicyEntries, PwPolicyManager
from lib389.password_plugins import PasswordPlugins
from lib389.idm.account import Account
from lib389.cli_base import CustomHelpFormatter


def _args_to_attrs(args, arg_to_attr):
    attrs = {}
    for arg in vars(args):
        val = getattr(args, arg)
        if arg in arg_to_attr and val is not None:
            attrs[arg_to_attr[arg]] = val
    return attrs


def _get_policy_type(inst, dn=None):
    pwp_manager = PwPolicyManager(inst)
    if dn is None:
        return "Global Password Policy"
    elif pwp_manager.is_subtree_policy(dn):
        return "Subtree Policy"
    else:
        return "User Policy"


def _get_pw_policy(inst, targetdn, log, use_json=None):
    pwp_manager = PwPolicyManager(inst)
    policy_type = _get_policy_type(inst, targetdn)
    attr_list = list(pwp_manager.arg_to_attr.values())
    if "global" in policy_type.lower():
        targetdn = 'cn=config'
        policydn = targetdn
        basedn = targetdn
        attr_list.extend(['passwordisglobalpolicy', 'nsslapd-pwpolicy-local'])
        all_attrs = inst.config.get_attrs_vals_utf8(attr_list)
        attrs = {k: v for k, v in all_attrs.items() if len(v) > 0}
    else:
        policy = pwp_manager.get_pwpolicy_entry(targetdn)
        basedn = policy.get_basedn()
        policydn = policy.dn
        all_attrs = policy.get_attrs_vals_utf8(attr_list)
        attrs = {k: v for k, v in all_attrs.items() if len(v) > 0}
    if use_json:
        log.info(json.dumps({
                 "dn": ensure_str(policydn),
                 "targetdn": targetdn,
                 "type": "entry",
                 "pwp_type": policy_type,
                 "basedn": basedn,
                 "attrs": attrs}, indent=4))
    else:
        if "global" in policy_type.lower():
            response = "Global Password Policy: cn=config\n------------------------------------\n"
        else:
            response = "Local {} Policy for \"{}\": {}\n------------------------------------\n".format(policy_type, targetdn, policydn)
        for key, value in list(attrs.items()):
            if len(value) == 0:
                value = ""
            else:
                value = value[0]
            response += "{}: {}\n".format(key, value)
        log.info(response)


def _log_policy_result(log, entry, use_json=None):
    """
    Logs the result of creating/updating a password policy
    in json or text format.
    """
    status = entry.ensure_status
    if status == entry.ENSURE_UNCHANGED:
        message = 'Password policy is already up to date'
        ensure = "UNCHANGED"
    elif status == entry.ENSURE_UPDATED:
        message = 'Successfully updated password policy'
        ensure = "UPDATED"
    elif status == entry.ENSURE_ADDED:
        message = 'Successfully created new password policy'
        ensure = "ADDED"
    else:
        message = "Unknown password policy operation"
        ensure = "UNKNOWN"

    if use_json:
        log.info(json.dumps({"ensure_status": ensure, "message": message}, indent=4))
    else:
        log.info(message)


def list_policies(inst, basedn, log, args):
    log = log.getChild('list_policies')

    if args.DN is None:
        # list all the password policies for all the backends
        targetdns = []
        backends = Backends(inst).list()
        for backend in backends:
            targetdns.append(backend.get_suffix())
    else:
        targetdns = [args.DN]

    if args.json:
        result = {'type': 'list', 'items': []}
    else:
        result = ""

    seen_dns = set()
    for targetdn in targetdns:
        # Verify target dn exists before getting started
        user_entry = Account(inst, targetdn)
        if not user_entry.exists():
            raise ValueError('The target entry dn does not exist')

        # User pwpolicy entry is under the container that is under the parent,
        # so we need to go one level up
        pwp_entries = PwPolicyEntries(inst, targetdn)
        pwp_manager = PwPolicyManager(inst)
        attr_list = list(pwp_manager.arg_to_attr.values())

        for pwp_entry in pwp_entries.list():
            # Filter duplicates because subtree search on parent suffix also returns
            # policies from sub suffixes
            if pwp_entry.dn in seen_dns:
                continue
            seen_dns.add(pwp_entry.dn)
            # Sometimes, the cn value includes quotes (for example, after migration from pre-CLI version).
            # We need to strip them as python-ldap doesn't expect them
            dn_comps_str = pwp_entry.get_attr_val_utf8_l('cn').strip("\'").strip("\"")
            try:
                dn_comps = ldap.dn.explode_dn(dn_comps_str)
                dn_comps.pop(0)
                entrydn = ",".join(dn_comps)
                policy_type = _get_policy_type(inst, entrydn)
            except ldap.DECODING_ERROR:
                # This is some kind of custom password policy, the UI relies on
                # on type being "Unknown policy type" in this.unknownPolicyType
                policy_type = "Unknown policy type"
                entrydn = "Unknown target"
                if not args.json:
                    # If not JSON then set a custom response, otherwise the
                    # JSON result is sufficient
                    result += "%s (%s)\n" % (pwp_entry.dn, policy_type.lower())
                    continue

            all_attrs = pwp_entry.get_attrs_vals_utf8(attr_list)
            attrs = {k: v for k, v in all_attrs.items() if len(v) > 0}
            if args.json:
                result['items'].append(
                    {
                        "dn": pwp_entry.dn,
                        "targetdn": entrydn,
                        "pwp_type": policy_type,
                        "basedn": pwp_entry.get_basedn(),
                        "attrs": attrs
                    }
                )
            else:
                result += "%s (%s)\n" % (entrydn, policy_type.lower())

    if args.json:
        log.info(json.dumps(result, indent=4))
    else:
        log.info(result)


def get_local_policy(inst, basedn, log, args):
    log = log.getChild('get_local_policy')
    _get_pw_policy(inst, args.DN[0], log, args.json)


def get_global_policy(inst, basedn, log, args):
    log = log.getChild('get_global_policy')
    _get_pw_policy(inst, None, log, args.json)


def create_subtree_policy(inst, basedn, log, args):
    log = log.getChild('create_subtree_policy')
    # Gather the attributes
    pwp_manager = PwPolicyManager(inst)
    attrs = _args_to_attrs(args, pwp_manager.arg_to_attr)
    pwp_entry = pwp_manager.create_subtree_policy(args.DN[0], attrs)
    _log_policy_result(log, pwp_entry, args.json)


def create_user_policy(inst, basedn, log, args):
    log = log.getChild('create_user_policy')
    pwp_manager = PwPolicyManager(inst)
    attrs = _args_to_attrs(args, pwp_manager.arg_to_attr)
    pwp_entry = pwp_manager.create_user_policy(args.DN[0], attrs)
    _log_policy_result(log, pwp_entry, args.json)


def set_global_policy(inst, basedn, log, args):
    log = log.getChild('set_global_policy')
    pwp_manager = PwPolicyManager(inst)
    attrs = _args_to_attrs(args, pwp_manager.arg_to_attr)
    pwp_manager.set_global_policy(attrs)
    log.info('Successfully updated global password policy')


def set_local_policy(inst, basedn, log, args):
    log = log.getChild('set_local_policy')
    targetdn = args.DN[0]
    pwp_manager = PwPolicyManager(inst)
    attrs = _args_to_attrs(args, pwp_manager.arg_to_attr)
    pwp_entry = pwp_manager.get_pwpolicy_entry(args.DN[0])
    policy_type = _get_policy_type(inst, targetdn)

    modlist = []
    for attr, value in attrs.items():
        modlist.append((attr, value))
    if len(modlist) > 0:
        pwp_entry.replace_many(*modlist)
    else:
        raise ValueError("There are no password policies to set")

    log.info('Successfully updated %s' % policy_type.lower())


def del_local_policy(inst, basedn, log, args):
    log = log.getChild('del_local_policy')
    targetdn = args.DN[0]
    policy_type = _get_policy_type(inst, targetdn)
    pwp_manager = PwPolicyManager(inst)
    pwp_manager.delete_local_policy(targetdn)
    log.info('Successfully deleted %s' % policy_type.lower())


def list_schemes(inst, basedn, log, args):
    schemes = PasswordPlugins(inst).list()
    scheme_list = []
    for scheme in schemes:
        scheme_list.append(scheme.get_attr_val_utf8('cn'))

    if args.json:
        result = {'type': 'list', 'items': scheme_list}
        log.info(json.dumps(result, indent=4))
    else:
        for scheme in scheme_list:
            log.info(scheme)


def fixup_shadow_last_change(inst, basedn, log, args):
    """Create the fixup shadow attributes task (shadowLastChange) and wait for it."""
    log = log.getChild('fixup_shadow_last_change')
    from lib389.tasks import ShadowFixupTask

    if not args.watch:
        log.info('Adding shadowLastChange fixup task for suffix "%s"%s ...',
                 args.suffix, ' (force)' if args.force else '')

    fixup_task = ShadowFixupTask(inst)
    fixup_task.create(args.suffix, force=args.force)
    if args.watch:
        fixup_task.watch()
    else:
        fixup_task.wait()
    result = fixup_task.get_exit_code()
    warning = fixup_task.get_task_warn()

    if result != 0:
        raise ValueError(f'shadowLastChange fixup task {fixup_task.dn} exited with code {result}')
    if warning:
        log.warning('shadowLastChange fixup task completed with warning code %s', warning)

    log.info('ShadowLastChange fixup task completed successfully')


def create_parser(subparsers):
    # Create our two parsers for local and global policies
    globalpwp_parser = subparsers.add_parser('pwpolicy', help='Manage the global password policy settings', formatter_class=CustomHelpFormatter)
    localpwp_parser = subparsers.add_parser('localpwp', help='Manage the local user and subtree password policies', formatter_class=CustomHelpFormatter)

    ############################################
    # Local password policies
    ############################################
    local_subcommands = localpwp_parser.add_subparsers(help='Local password policy')
    # List all the local policies
    list_parser = local_subcommands.add_parser('list', help='List all the local password policies', formatter_class=CustomHelpFormatter)
    list_parser.set_defaults(func=list_policies)
    list_parser.add_argument('DN', nargs='?', help='Suffix to search for local password policies')
    # Get a local policy
    get_parser = local_subcommands.add_parser('get', help='Get local password policy entry', formatter_class=CustomHelpFormatter)
    get_parser.set_defaults(func=get_local_policy)
    get_parser.add_argument('DN', nargs=1, help='Get the local policy for this entry DN')
    # The "set" arguments...
    set_parser = local_subcommands.add_parser('set', help='Set an attribute in a local password policy', formatter_class=CustomHelpFormatter)
    set_parser.set_defaults(func=set_local_policy)
    # General settings
    set_parser.add_argument('--pwdscheme', help="The password storage scheme")
    set_parser.add_argument('--pwdchange', help="Allow users to change their passwords")
    set_parser.add_argument('--pwdmustchange', help="Users must change their password after it was reset by an administrator")
    set_parser.add_argument('--pwdhistory', help="To enable password history set this to \"on\", otherwise \"off\"")
    set_parser.add_argument('--pwdhistorycount', help="The number of passwords to keep in history")
    set_parser.add_argument('--pwdadmin', help="The DN of an entry or a group of account that can bypass password policy constraints")
    set_parser.add_argument('--pwdadminskipupdates',
                            help="Set to \"on\" if the Password Admin's password update should not trigger updates to the password state attributes (passwordExpirationtime, passwordHistory, etc).")

    set_parser.add_argument('--pwdtrack', help="Set to \"on\" to track the time the password was last changed")
    set_parser.add_argument('--pwdwarning', help="Send an expiring warning if password expires within this time (in seconds)")
    # Expiration settings
    set_parser.add_argument('--pwdexpire', help="Set to \"on\" to enable password expiration")
    set_parser.add_argument('--pwdmaxage', help="The password expiration time in seconds")
    set_parser.add_argument('--pwdminage', help="The number of seconds that must pass before a user can change their password")
    set_parser.add_argument('--pwdgracelimit', help="The number of allowed logins after the password has expired")
    set_parser.add_argument('--pwdsendexpiring', help="Set to \"on\" to always send the expiring control regardless of the warning period")
    # Account lockout settings
    set_parser.add_argument('--pwdlockout', help="Set to \"on\" to enable account lockout")
    set_parser.add_argument('--pwdunlock', help="Set to \"on\" to allow an account to become unlocked after the lockout duration")
    set_parser.add_argument('--pwdlockoutduration', help="The number of seconds an account stays locked out")
    set_parser.add_argument('--pwdmaxfailures', help="The maximum number of allowed failed password attempts before the account gets locked")
    set_parser.add_argument('--pwdresetfailcount', help="The number of seconds to wait before reducing the failed login count on an account")
    # Syntax settings
    set_parser.add_argument('--pwdchecksyntax', help="Set to \"on\" to enable password syntax checking")
    set_parser.add_argument('--pwdminlen', help="The minimum number of characters required in a password")
    set_parser.add_argument('--pwdmindigits', help="The minimum number of digit/number characters in a password")
    set_parser.add_argument('--pwdminalphas', help="The minimum number of alpha characters required in a password")
    set_parser.add_argument('--pwdminuppers', help="The minimum number of uppercase characters required in a password")
    set_parser.add_argument('--pwdminlowers', help="The minimum number of lowercase characters required in a password")
    set_parser.add_argument('--pwdminspecials', help="The minimum number of special characters required in a password")
    set_parser.add_argument('--pwdmin8bits', help="The minimum number of 8-bit characters required in a password")
    set_parser.add_argument('--pwdmaxrepeats', help="The maximum number of times the same character can appear sequentially in the password")
    set_parser.add_argument('--pwdpalindrome', help="Set to \"on\" to reject passwords that are palindromes")
    set_parser.add_argument('--pwdmaxseq', help="The maximum number of allowed monotonic character sequences in a password")
    set_parser.add_argument('--pwdmaxseqsets', help="The maximum number of allowed monotonic character sequences that can be duplicated in a password")
    set_parser.add_argument('--pwdmaxclasschars', help="The maximum number of sequential characters from the same character class that is allowed in a password")
    set_parser.add_argument('--pwdmincatagories', help="The minimum number of syntax category checks")
    set_parser.add_argument('--pwdmintokenlen', help="Sets the smallest attribute value length that is used for trivial/user words checking.  This also impacts \"--pwduserattrs\"")
    set_parser.add_argument('--pwdbadwords', help="A space-separated list of words that can not be in a password")
    set_parser.add_argument('--pwduserattrs', help="A space-separated list of attributes whose values can not appear in the password (See \"--pwdmintokenlen\")")
    set_parser.add_argument('--pwddictcheck', help="Set to \"on\" to enforce CrackLib dictionary checking")
    set_parser.add_argument('--pwddictpath', help="Filesystem path to specific/custom CrackLib dictionary files")
    set_parser.add_argument('--pwptprmaxuse', help="Number of times a reset password can be used for authentication")
    set_parser.add_argument('--pwptprdelayexpireat', help="Number of seconds after which a reset password expires")
    set_parser.add_argument('--pwptprdelayvalidfrom', help="Number of seconds to wait before using a reset password to authenticated")
    # delete local password policy
    del_parser = local_subcommands.add_parser('remove', help='Remove a local password policy', formatter_class=CustomHelpFormatter)
    del_parser.set_defaults(func=del_local_policy)
    del_parser.add_argument('DN', nargs=1, help='Remove local policy for this entry DN')
    #
    # create USER local password policy
    #
    add_user_parser = local_subcommands.add_parser('adduser', add_help=False, parents=[set_parser], help='Add new user password policy', formatter_class=CustomHelpFormatter)
    add_user_parser.set_defaults(func=create_user_policy)
    #
    # create SUBTREE local password policy
    #
    add_subtree_parser = local_subcommands.add_parser('addsubtree', add_help=False, parents=[set_parser], help='Add new subtree password policy', formatter_class=CustomHelpFormatter)
    add_subtree_parser.set_defaults(func=create_subtree_policy)

    ###########################################
    # The global policy (cn=config)
    ###########################################
    global_subcommands = globalpwp_parser.add_subparsers(help='Global password policy')
    # Get policy
    get_global_parser = global_subcommands.add_parser('get', help='Get the global password policy entry', formatter_class=CustomHelpFormatter)
    get_global_parser.set_defaults(func=get_global_policy)
    # Set policy
    set_global_parser = global_subcommands.add_parser('set', add_help=False, parents=[set_parser],
                                                      help='Set an attribute in a global password policy')
    set_global_parser.set_defaults(func=set_global_policy)
    set_global_parser.add_argument('--pwdlocal', help="Set to \"on\" to enable fine-grained (subtree/user-level) password policies")
    set_global_parser.add_argument('--pwdisglobal', help="Set to \"on\" to enable password policy state attributes to be replicated")
    set_global_parser.add_argument('--pwdallowhash', help="Set to \"on\" to allow adding prehashed passwords")
    set_global_parser.add_argument('--pwpinheritglobal', help="Set to \"on\" to allow local policies to inherit the global policy")
    # list password storage schemes
    list_scehmes_parser = global_subcommands.add_parser('list-schemes', help='Get a list of the current password storage schemes', formatter_class=CustomHelpFormatter)
    list_scehmes_parser.set_defaults(func=list_schemes)
    # Fix up shadowLastChange on ShadowAccount entries under a suffix
    fixup_shadow_parser = global_subcommands.add_parser(
        'fixup-shadow',
        help='Create a task to set shadowLastChange on ShadowAccount entries',
        formatter_class=CustomHelpFormatter)
    fixup_shadow_parser.set_defaults(func=fixup_shadow_last_change)
    fixup_shadow_parser.add_argument(
        'suffix', help='LDAP suffix to search (subtree) for objectClass ShadowAccount')
    fixup_shadow_parser.add_argument(
        '--force', action='store_true',
        help='Update all users under the suffix regardless if shadowLastChange is set.')
    fixup_shadow_parser.add_argument(
        '--watch', action='store_true',
        help='Watch the task\'s status and wait for it to finish.')

    #############################################
    # Wrap it up.  Now that we copied all the parent arguments to the subparsers,
    # and the DN argument needed only by the local policies
    #############################################
    set_parser.add_argument('DN', nargs=1, help='Set the local policy for this entry DN')
    add_subtree_parser.add_argument('DN', nargs=1, help='Add/replace the subtree policy for this entry DN')
    add_user_parser.add_argument('DN', nargs=1, help='Add/replace the local password policy for this entry DN')



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.plugins import LinkedAttributesPlugin, LinkedAttributesConfig, LinkedAttributesConfigs, LinkedAttributesFixupTasks
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add
from lib389.cli_base import CustomHelpFormatter
from lib389.utils import get_task_status

arg_to_attr = {
    'link_type': 'linkType',
    'managed_type': 'managedType',
    'link_scope': 'linkScope',
}


def linkedattr_list(inst, basedn, log, args):
    log = log.getChild('linkedattr_list')
    configs = LinkedAttributesConfigs(inst)
    result = []
    result_json = []
    for config in configs.list():
        if args.json:
            result_json.append(json.loads(config.get_all_attrs_json()))
        else:
            result.append(config.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Linked Attributes plugin config instances")


def linkedattr_add(inst, basedn, log, args):
    log = log.getChild('linkedattr_add')
    plugin = LinkedAttributesPlugin(inst)
    props = {'cn': args.NAME}
    generic_object_add(LinkedAttributesConfig, inst, log, args, arg_to_attr, basedn=plugin.dn, props=props)


def linkedattr_edit(inst, basedn, log, args):
    log = log.getChild('linkedattr_edit')
    configs = LinkedAttributesConfigs(inst)
    config = configs.get(args.NAME)
    generic_object_edit(config, log, args, arg_to_attr)


def linkedattr_show(inst, basedn, log, args):
    log = log.getChild('linkedattr_show')
    configs = LinkedAttributesConfigs(inst)
    config = configs.get(args.NAME)

    if not config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.name)
    if args and args.json:
        o_str = config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(config.display())


def linkedattr_del(inst, basedn, log, args):
    log = log.getChild('linkedattr_del')
    configs = LinkedAttributesConfigs(inst)
    config = configs.get(args.NAME)
    config.delete()
    log.info("Successfully deleted the %s", config.dn)


def fixup(inst, basedn, log, args):
    plugin = LinkedAttributesPlugin(inst)
    log.info('Attempting to add task entry... This will fail if LinkedAttributes plug-in is not enabled.')
    if not plugin.status():
        log.error("'%s' is disabled. Fix up task can't be executed" % plugin.rdn)
    fixup_task = plugin.fixup(args.linkdn)

    if args.wait:
        log.info(f'Waiting for fixup task "{fixup_task.dn}" to complete.  You can safely exit by pressing Control C ...')
        fixup_task.wait(timeout=None)
        exitcode = fixup_task.get_exit_code()
        if exitcode != 0:
            log.error(f'LinkedAttributes fixup "{fixup_task.dn}" for {args.linkdn} has failed (error {exitcode}). Please, check logs')
        else:
            log.info('Fixup task successfully completed')
    else:
        log.info(f'Successfully added task entry "{fixup_task.dn}". This task is running in the background. To track its progress you can use the "fixup-status" command.')


def do_fixup_status(inst, basedn, log, args):
    get_task_status(inst, log, LinkedAttributesFixupTasks, dn=args.dn, show_log=args.show_log,
                    watch=args.watch, use_json=args.json)


def _add_parser_args(parser):
    parser.add_argument('--link-type',
                        help='Sets the attribute that is managed manually by administrators (linkType)')
    parser.add_argument('--managed-type',
                        help='Sets the attribute that is created dynamically by the plugin (managedType)')
    parser.add_argument('--link-scope',
                        help='Sets the scope that restricts the plugin to a specific part of the directory tree (linkScope)')


def create_parser(subparsers):
    linkedattr_parser = subparsers.add_parser('linked-attr', help='Manage and configure Linked Attributes plugin', formatter_class=CustomHelpFormatter)
    subcommands = linkedattr_parser.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, LinkedAttributesPlugin)

    fixup_parser = subcommands.add_parser('fixup', help='Run the fix-up task for linked attributes plugin', formatter_class=CustomHelpFormatter)
    fixup_parser.add_argument('-l', '--linkdn', help="Sets the base DN that contains entries to fix up")
    fixup_parser.add_argument('--wait', action='store_true',
                              help="Wait for the task to finish, this could take a long time")
    fixup_parser.set_defaults(func=fixup)

    fixup_status = subcommands.add_parser('fixup-status', help='Check the status of a fix-up task', formatter_class=CustomHelpFormatter)
    fixup_status.set_defaults(func=do_fixup_status)
    fixup_status.add_argument('--dn', help="The task entry's DN")
    fixup_status.add_argument('--show-log', action='store_true', help="Display the task log")
    fixup_status.add_argument('--watch', action='store_true',
                       help="Watch the task's status and wait for it to finish")

    list = subcommands.add_parser('list', help='List available plugin configs', formatter_class=CustomHelpFormatter)
    list.set_defaults(func=linkedattr_list)

    config = subcommands.add_parser('config', help='Manage plugin configs', formatter_class=CustomHelpFormatter)
    config.add_argument('NAME', help='The Linked Attributes configuration name')
    config_subcommands = config.add_subparsers(help='action')
    add = config_subcommands.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=linkedattr_add)
    _add_parser_args(add)
    edit = config_subcommands.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=linkedattr_edit)
    _add_parser_args(edit)
    show = config_subcommands.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show.set_defaults(func=linkedattr_show)
    delete = config_subcommands.add_parser('delete', help='Delete the config entry', formatter_class=CustomHelpFormatter)
    delete.set_defaults(func=linkedattr_del)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 William Brown <william@blackhats.net.au>
# Copyright (C) 2024 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.plugins import Plugin, Plugins
from lib389._constants import DN_PWDSTORAGE_SCHEMES


class PasswordPlugin(Plugin):
    _plugin_properties = {
        'nsslapd-pluginpath': 'libpwdstorage-plugin',
        'nsslapd-plugintype': 'pwdstoragescheme',
        'nsslapd-pluginEnabled' : 'on'
    }

    def __init__(self, instance, dn=None):
        super(PasswordPlugin, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = [
            'nsslapd-pluginEnabled',
            'nsslapd-pluginPath',
            'nsslapd-pluginInitfunc',
            'nsslapd-pluginType',
            ]
        self._create_objectclasses = ['top', 'nsslapdplugin']
        # We'll mark this protected, and people can just disable the plugins.
        self._protected = True


class SSHA512Plugin(PasswordPlugin):
    def __init__(self, instance, dn=f'cn=SSHA512,{DN_PWDSTORAGE_SCHEMES}'):
        super(SSHA512Plugin, self).__init__(instance, dn)


class SHAPlugin(PasswordPlugin):
    def __init__(self, instance, dn=f'cn=SHA,{DN_PWDSTORAGE_SCHEMES}'):
        super(SHAPlugin, self).__init__(instance, dn)


class CRYPTPlugin(PasswordPlugin):
    def __init__(self, instance, dn=f'cn=CRYPT,{DN_PWDSTORAGE_SCHEMES}'):
        super(CRYPTPlugin, self).__init__(instance, dn)


class SSHAPlugin(PasswordPlugin):
    def __init__(self, instance, dn=f'cn=SSHA,{DN_PWDSTORAGE_SCHEMES}'):
        super(SSHAPlugin, self).__init__(instance, dn)


class PBKDF2BasePlugin(PasswordPlugin):
    """Base class for all PBKDF2 variants"""
    DEFAULT_ROUNDS = 100000
    ACCEPT_MAX_DEFAULT = 10000000
    ACCEPT_MAX_MIN = 10000
    ACCEPT_MAX_MAX = 10000000

    def __init__(self, instance, dn):
        super(PBKDF2BasePlugin, self).__init__(instance, dn)
        self._create_objectclasses.append('pwdPBKDF2PluginConfig')

    def set_rounds(self, rounds):
        """Set the number of rounds for PBKDF2 hashing (requires restart)

        :param rounds: Number of rounds
        :type rounds: int
        """
        # Ensure the pwdPBKDF2PluginConfig objectClass is present
        self.ensure_present('objectClass', 'pwdPBKDF2PluginConfig')

        rounds = int(rounds)
        if rounds < 10000 or rounds > 10000000:
            raise ValueError("PBKDF2 rounds must be between 10,000 and 10,000,000")
        self.replace('nsslapd-pwdPBKDF2NumIterations', str(rounds))

    def get_rounds(self):
        """Get the current number of rounds

        :param use_json: Whether to return JSON formatted output
        :type use_json: bool
        :returns: Current rounds setting or JSON string
        :rtype: int
        """
        rounds = self.get_attr_val_int('nsslapd-pwdPBKDF2NumIterations')
        if rounds:
            return rounds
        return self.DEFAULT_ROUNDS

    def set_accept_max_iterations(self, accept_max):
        """Set the maximum PBKDF2 iterations accepted from stored hashes on bind (requires restart).

        :param accept_max: Maximum iteration count accepted on compare
        :type accept_max: int
        """
        self.ensure_present('objectClass', 'pwdPBKDF2PluginConfig')

        accept_max = int(accept_max)
        if accept_max < self.ACCEPT_MAX_MIN or accept_max > self.ACCEPT_MAX_MAX:
            raise ValueError(
                "PBKDF2 accept max iterations must be between 10,000 and 10,000,000"
            )
        self.replace('nsslapd-pwdPBKDF2AcceptMaxIterations', str(accept_max))

    def get_accept_max_iterations(self):
        """Get the maximum PBKDF2 iterations accepted from stored hashes on bind.

        :returns: Configured accept max, or default if unset
        :rtype: int
        """
        accept_max = self.get_attr_val_int('nsslapd-pwdPBKDF2AcceptMaxIterations')
        if accept_max:
            return accept_max
        return self.ACCEPT_MAX_DEFAULT

    def delete_accept_max_iterations(self):
        """Remove the accept max iterations override, revert to default."""
        self.ensure_present('objectClass', 'pwdPBKDF2PluginConfig')
        self.remove_all('nsslapd-pwdPBKDF2AcceptMaxIterations')


class PBKDF2Plugin(PBKDF2BasePlugin):
    """PBKDF2 password storage scheme"""

    def __init__(self, instance, dn=f'cn=PBKDF2,{DN_PWDSTORAGE_SCHEMES}'):
        super(PBKDF2Plugin, self).__init__(instance, dn)


class PBKDF2SHA1Plugin(PBKDF2BasePlugin):
    """PBKDF2-SHA1 password storage scheme"""

    def __init__(self, instance, dn=f'cn=PBKDF2-SHA1,{DN_PWDSTORAGE_SCHEMES}'):
        super(PBKDF2SHA1Plugin, self).__init__(instance, dn)


class PBKDF2SHA256Plugin(PBKDF2BasePlugin):
    """PBKDF2-SHA256 password storage scheme"""

    def __init__(self, instance, dn=f'cn=PBKDF2-SHA256,{DN_PWDSTORAGE_SCHEMES}'):
        super(PBKDF2SHA256Plugin, self).__init__(instance, dn)


class PBKDF2SHA512Plugin(PBKDF2BasePlugin):
    """PBKDF2-SHA512 password storage scheme"""

    def __init__(self, instance, dn=f'cn=PBKDF2-SHA512,{DN_PWDSTORAGE_SCHEMES}'):
        super(PBKDF2SHA512Plugin, self).__init__(instance, dn)


class PBKDF2SHA256LegacyPlugin(PasswordPlugin):
    """Legacy C PBKDF2_SHA256 password storage scheme"""
    ACCEPT_MAX_MIN = 2048
    ACCEPT_MAX_MAX = 2147483647
    ACCEPT_MAX_DEFAULT = 50000

    def __init__(self, instance, dn=f'cn=PBKDF2_SHA256,{DN_PWDSTORAGE_SCHEMES}'):
        super(PBKDF2SHA256LegacyPlugin, self).__init__(instance, dn)
        self._create_objectclasses.append('pwdPBKDF2PluginConfig')

    def set_accept_max_iterations(self, iterations):
        """Set the maximum accepted PBKDF2 iteration count (requires restart)

        :param iterations: Maximum accepted iteration count
        :type iterations: int
        """
        self.ensure_present('objectClass', 'pwdPBKDF2PluginConfig')

        iterations = int(iterations)
        if iterations < self.ACCEPT_MAX_MIN or iterations > self.ACCEPT_MAX_MAX:
            raise ValueError(
                "PBKDF2 accept max iterations must be between 2,048 and 2,147,483,647"
            )
        self.replace('nsslapd-pwdPBKDF2AcceptMaxIterations', str(iterations))

    def get_accept_max_iterations(self):
        """Get the maximum accepted PBKDF2 iteration count

        :returns: Maximum accepted iteration count
        :rtype: int
        """
        iterations = self.get_attr_val_int('nsslapd-pwdPBKDF2AcceptMaxIterations')
        if iterations:
            return iterations
        return self.ACCEPT_MAX_DEFAULT


class PasswordPlugins(Plugins):
    def __init__(self, instance):
        super(PasswordPlugins, self).__init__(instance=instance)
        self._objectclasses = ['nsSlapdPlugin']
        self._filterattrs = ['cn']
        self._childobject = PasswordPlugin
        self._basedn = DN_PWDSTORAGE_SCHEMES



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

#
# This module handles dsconf libdb sub-commands
#

import sys
import os
import glob
import pwd
import re
import logging
import shutil
import subprocess
from enum import Enum
from errno import ENOSPC
from lib389.cli_base import CustomHelpFormatter
from lib389._constants import DEFAULT_LMDB_SIZE, BDB_IMPL_STATUS, DN_CONFIG, DBSCAN
from lib389.dseldif import DSEldif
from lib389.utils import parse_size, format_size, check_plugin_strings, find_plugin_path
from lib389.paths import Paths
from pathlib import Path


DBLIB_LDIF_PREFIX = "__dblib-"
DBSIZE_MARGIN = 1.2
DBI_MARGIN = 60

MDB_INFO = "INFO.mdb"
MDB_MAP = "data.mdb"
MDB_LOCK = "lock.mdb"

LDBM_DN = "cn=config,cn=ldbm database,cn=plugins,cn=config"
CL5DB='replication_changelog.db'

_log = None


class FakeArgs(dict):
    # This calss is used for tests to generate an args
    # It allows to access a dict key as an attribute
    # (be cautious not to use a key that is a dict attibute like (keys, items, ... )
    def __init__(self, *args, **kwargs):
        super(FakeArgs, self).__init__(*args, **kwargs)
        self.__dict__ = self


class CalledProcessUnexpectedReturnCode(subprocess.CalledProcessError):
    def __init__(self, result, expected_rc):
        super().__init__(cmd=result.args, returncode=result.returncode, output=result.stdout, stderr=result.stderr)
        self.expected_rc = expected_rc
        self.result = result

    def __str__(self):
        return f'Command {self.result.args} returned {self.result.returncode} instead of {self.expected_rc}'


class DbscanHelper:
    """
    Helper class to get the existing database(s) dbis
    Usage: DbscanHelper(inst).get_dbi('attrname' [, backend='backendname'])
           throws KeyError if there is no associated dbi
    """

    def __init__(self, inst, log=None):
        self.inst = inst
        self.log = log if log else logging.getLogger(__name__)

    def _init2(self):
        if getattr(self, 'dblib', None) is not None:
            return
        assert(self.inst)
        self.dblib = self.inst.get_db_lib()
        self.dbhome = self.inst.ds_paths.db_dir
        self.dbis = self._list_instances()
        self.ldif_dir = self.inst.ds_paths.ldif_dir

        # set self.options with dbscan supported options
        self.options = []
        usage = self.dbscan(['-h'], expected_rc=None).stdout
        pattern = r'^\s+(?:(-[^-,]+), +)?(--[^ ]+).*$'
        for match in re.finditer(pattern, usage, flags=re.MULTILINE):
            for idx in range(1,3):
                if match.group(idx) is not None:
                    self.options.append(match.group(idx))

    def resync(self):
        assert(self.inst)
        self.dblib = None
        self._init2()

    def _list_instances(self):
        # compute db instance pathnames
        instances = self.dbscan(['-D', self.dblib, '-L', self.dbhome]).stdout
        dbis = []
        if self.dblib == 'bdb':
            pattern = r'^ (.*) $'
        else:
            pattern = r'^ (.*) flags:'
        for match in re.finditer(pattern, instances, flags=re.MULTILINE):
            dbis.append(match.group(1))
        return dbis

    def get_dbi(self, attr, backend='userroot'):
        self._init2()
        for dbi in self.dbis:
            if f'{backend}/{attr}.'.lower() in dbi.lower():
                return dbi
        raise KeyError(f'Unknown dbi {backend}/{attr}')

    def dbscan(self, args, expected_rc=0):
        if self.inst is None:
            prefix = os.environ.get('PREFIX', "")
            prog = f'{prefix}/bin/dbscan'
        else:
            prog = os.path.join(self.inst.ds_paths.bin_dir, DBSCAN)
        args.insert(0, prog)
        output = subprocess.run(args, encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.log.debug(f'{args} result is {output.returncode} output is {output.stdout}')
        if expected_rc is not None and expected_rc != output.returncode:
            raise CalledProcessUnexpectedReturnCode(output, expected_rc)
        return output

    def __repr__(self):
        if self.inst is None:
            return 'DbscanHelper(None)'
        self._init2()
        attrs = ['inst', 'dblib', 'dbhome', 'ldif_dir', 'options', 'dbis' ]
        res = ", ".join(map(lambda x: f'{x}={getattr(self, x, None)}', attrs))
        return f'DbscanHelper({res})'


def get_bdb_impl_status():
    p = Paths()
    p._read_defaults()
    libdir = f"{p._config['slapd']['lib_dir']}/dirsrv"
    robdb = glob.glob(f'{libdir}/librobdb.so*')
    has_robdb = len(robdb) > 0
    backldbm = 'libback-ldbm'
    bundledbdb_plugin = 'libback-bdb'
    libdb = 'libdb-'
    plgstrs = check_plugin_strings(backldbm, [bundledbdb_plugin, libdb])
    has_bundled_strings = plgstrs[bundledbdb_plugin] is True
    has_standard_strings = plgstrs[libdb] is True

    # Check read-only BDB
    if has_robdb:
        return BDB_IMPL_STATUS.READ_ONLY

    # Check bundled BDB
    if has_bundled_strings and find_plugin_path(bundledbdb_plugin):
        return BDB_IMPL_STATUS.BUNDLED

    # Check standard (provided by system) BDB
    if has_standard_strings:
        return BDB_IMPL_STATUS.STANDARD

    # If bundled strings found but no working implementation
    if has_bundled_strings:
        return BDB_IMPL_STATUS.NONE

    # Unable to find any BDB indicators in libback-ldbm plugin
    return BDB_IMPL_STATUS.UNKNOWN

def is_bdb_supported(read_write=True):
    bdbok = [BDB_IMPL_STATUS.BUNDLED, BDB_IMPL_STATUS.STANDARD]
    if not read_write:
        # READ_MODE is ok too
        bdbok.append(BDB_IMPL_STATUS.READ_ONLY)
    return get_bdb_impl_status() in bdbok


def get_ldif_dir(instance):
    """
    Get the server's LDIF directory.
    """
    server_dir = instance.get_ldif_dir()
    if server_dir is not None:
        return server_dir
    return "/foo"


def get_backends(log, dse, tmpdir):
    """
    Get the backends and some associated data
    Note: inst is not connected so Backends(inst).list() cannot be used.
          So lets directly parse the data from the ldif file
    """
    res = {}
    ic2ec = {}
    # Lets be sure to keep config backend info
    ic2ec['config'] = 'config'
    update_dse = []
    dbis = None
    for entry in dse._contents:
        found = re.search(r'^nsslapd-backend: (.*)', entry)
        if found:
            ic2ec[found[1].lower()] = found[1]
        found = re.search(r'^dn: cn=([^,]*), *cn=ldbm database.*', entry)
        if found:
            dn = found[0][4:]
            bename = found[1].lower()
            if not bename in ic2ec:
                # Not a mapping tree backend
                continue
            ecbename = ic2ec[bename]
            suffix = dse.get(dn, "nsslapd-suffix", True)
            dbdir = dse.get(dn, "nsslapd-directory", True)
            ecdbdir = dbdir
            dblib = dse.get(dn, "nsslapd-backend-implement", True)
            if dblib is None and 'config' in res:
                dblib = res['config']['dblib']
            if dbis is None:
                dbis = get_mdb_dbis(dbdir) if dblib == 'mdb' else []
            # in bdb case dbdir should be the database instance directory
            # in mdb case it is the database map file directory
            if 'config' in res and ( dbdir is None or dblib == 'mdb'):
                dbdir = f"{res['config']['dbdir']}/{bename}"
                ecdbdir = f"{res['config']['dbdir']}/{ecbename}"
                # bdb requires nsslapd-directory so lets add it once reading is done.
                update_dse.append((dn, ecdbdir))
            ldifname = f'{tmpdir}/{DBLIB_LDIF_PREFIX}{bename}.ldif'
            cl5name = f'{tmpdir}/{DBLIB_LDIF_PREFIX}{bename}.cl5.dbtxt'
            cl5dbname = f'{dbdir}/{CL5DB}'
            eccl5dbname = f'{ecdbdir}/{CL5DB}'
            dbsize = 0
            entrysize = 0
            for f in glob.glob(f'{dbdir}/id2entry.db*'):
                entrysize = os.path.getsize(f)
            for f in glob.glob(f'{dbdir}/*.db*'):
                dbsize += os.path.getsize(f)
            indexes = dse.get_indexes(bename)
            # Let estimate the number of dbis: id2entry + 1 per regular index + 2 per vlv index
            dbi = 1 + len(indexes) + len([index for index in indexes if index.startswith("vlv#")])
            if dblib == "bdb":
                has_changelog = os.path.isfile(eccl5dbname)
            elif bename in dbis:
                has_changelog = CL5DB in dbis[bename]
            else:
                has_changelog = False

            res[bename] = {
                'dn': dn,
                'bename': bename,
                'ecbename': ecbename,
                'suffix': suffix,
                'dbdir': dbdir,
                'ecdbdir': ecdbdir,
                'dbsize': dbsize,
                'dblib': dblib,
                'ldifname': ldifname,
                'cl5name': cl5name,
                'cl5dbname': cl5dbname,
                'eccl5dbname': eccl5dbname,
                'dbsize': dbsize,
                'entrysize': entrysize,
                'indexes': indexes,
                'has_changelog': has_changelog,
                'dbi': dbi
            }

    # now that we finish reading the dse.ldif we may update it if needed.
    for dn, dir in update_dse:
        dse.replace(dn, 'nsslapd-directory', dir)
    log.debug(f'lib389.cli_ctl.dblib.get_backends returns: {str(res)}')
    return (res, dbis)


def get_mdb_dbis(dbdir):
    # Returns a map  containings associating the backend (Could be None)
    # to a map associating the dbi filename to a map describing its statistics.
    result = {}
    output = run_dbscan(['-D', 'mdb', '-L', dbdir])
    for line in output.split("\n"):
        found = re.search(f'^ {dbdir}/?([^/]*)?/([^/]*) flags: .*state:.([^ ]*).*nb_entries=([0-9]*)', line)
        if (found):
            bename = found[1].lower()
            if bename == '':
                bename = None
            filename = found[2]
            state = found[3]
            nbentries = found[4]
            if bename not in result:
                result[bename] = {}
            result[bename][filename] = {'bename': bename, 'filename': filename, 'state': state, 'nbentries': nbentries}
    return result


def run_dbscan(args):
    prefix = os.environ.get('PREFIX', "")
    prog = f'{prefix}/bin/dbscan'
    args.insert(0, prog)
    try:
        output = subprocess.check_output(args, encoding='utf-8', stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        _log.error(f"run_dbscan failed: cmd is {e.cmd} return code is {e.returncode} output is {str(e.output)}")
        _log.exception(e)
        raise e
    return output


def export_changelog(be, dblib):
    # Export backend changelog
    if not be['has_changelog']:
        return False
    try:
        cl5dbname = be['eccl5dbname'] if dblib == "bdb" else be['cl5dbname']
        _log.info(f"Exporting changelog {cl5dbname} to {be['cl5name']}")
        run_dbscan(['-D', dblib, '-f', cl5dbname, '-X', be['cl5name']])
        return True
    except subprocess.CalledProcessError as e:
        return False


def import_changelog(be, dblib):
    # import backend changelog
    try:
        cl5dbname = be['eccl5dbname'] if dblib == "bdb" else be['cl5dbname']
        _log.info(f"Importing changelog {cl5dbname} from {be['cl5name']}")
        run_dbscan(['-D', dblib, '-f', cl5dbname, '--import', be['cl5name'], '--do-it'])
        return True
    except subprocess.CalledProcessError as e:
        return False


def set_owner(list, uid, gid):
    global _log
    _log.debug(f"set_owner: {list} {uid} {gid}")
    for f in list:
        try:
            _log.debug(f"set_owner: {f} {uid} {gid}")
            os.chown(f, uid, gid)
        except OSError:
            pass


def get_uid_gid_from_dse(dse):
    # For some reason inst.get_user_uid() always returns dirsrv uid
    # even for non root install.
    # Lets looks directly in the dse.ldif intead
    user = dse.get(DN_CONFIG, 'nsslapd-localuser', single=True)
    pwent = pwd.getpwnam(user)
    return ( pwent.pw_uid, pwent.pw_gid )


def dblib_bdb2mdb(inst, log, args):
    global _log
    _log = log
    if args.tmpdir is None:
        tmpdir = get_ldif_dir(inst)
    else:
        tmpdir = args.tmpdir
    try:
        os.makedirs(tmpdir, 0o750, True)
    except OSError as e:
        log.error(f"Failed trying to create the directory {tmpdir} needed to store the ldif files, error: {str(e)}")
        return

    status = get_bdb_impl_status()
    if status is BDB_IMPL_STATUS.UNKNOWN:
        log.warning('Unable to determine if Berkeley Database library is available. If it is not the case, the migration will fail.')
    elif status is BDB_IMPL_STATUS.NONE:
        log.error('Berkeley Database library is not available. Maybe 389-ds-base-bdb rpm should be installed.')
        raise RuntimeError('Berkeley Database library is not available')

    # Cannot use Backends(inst).list() because it requires a connection.
    # lets use directlt the dse.ldif after having stopped the instance

    inst.stop()
    dse = DSEldif(inst)
    uid, gid = get_uid_gid_from_dse(dse)
    backends,dbis = get_backends(log, dse, tmpdir)
    dbmapdir = backends['config']['dbdir']
    dblib = backends['config']['dblib']

    if dblib == "mdb":
        log.error(f"Instance {inst.serverid} is already configured with lmdb.")
        return

    # Remove ldif files and mdb files
    dblib_cleanup(inst, log, args)

    # Check if changelog is present in backends

    # Compute the needed space and the lmdb map configuration
    total_dbsize = 0
    total_entrysize = 0
    total_dbi = 3
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if be['dbsize'] == 0:
            continue
        total_dbsize += be['dbsize']
        total_entrysize += be['entrysize']
        total_dbi += be['dbi']

    required_dbsize = round(total_dbsize * DBSIZE_MARGIN)

    # Compute a dbmap size greater than required_dbsize
    dbmap_size = parse_size(DEFAULT_LMDB_SIZE)
    while (required_dbsize > dbmap_size):
        dbmap_size = round(dbmap_size * 1.25)

    # Round up number of dbis
    nbdbis = 1
    while nbdbis < total_dbi + DBI_MARGIN:
        nbdbis *= 2

    log.info(f"Required space for LDIF files is about {format_size(total_entrysize)}")
    log.info(f"Required space for DBMAP files is about {format_size(required_dbsize)}")
    log.info(f"Required number of dbi is {nbdbis}")

    # Generate the info file (so dbscan could generate the map)
    with open(f'{dbmapdir}/{MDB_INFO}', 'w') as f:
        f.write('LIBVERSION=9025\n')
        f.write('DATAVERSION=0\n')
        f.write(f'MAXSIZE={dbmap_size}\n')
        f.write('MAXREADERS=50\n')
        f.write(f'MAXDBS={nbdbis}\n')
    os.chown(f'{dbmapdir}/{MDB_INFO}', uid, gid)

    total, used, free = shutil.disk_usage(dbmapdir)
    if os.stat(dbmapdir).st_dev != os.stat(tmpdir).st_dev:
        # Ldif and db are on different filesystems
        # Let check that we have enough space in tmpdir for ldif files
        total, used, free = shutil.disk_usage(tmpdir)
        if free < total_entrysize:
            raise OSError(ENOSPC, "Not enough space on {tmpdir} to migrate to lmdb " +
                                  "(In {tmpdir}, {format_size(total_entrysize)} is "+
                                  "needed but only {format_size(free)} is available)")
        total_entrysize = 0    # do not count total_entrysize when checking dbmapdir size

    # Let check that we have enough space in dbmapdir for the db and ldif files
    total, used, free = shutil.disk_usage(dbmapdir)
    size = required_dbsize + total_entrysize
    if free < required_dbsize + total_entrysize:
            raise OSError(ENOSPC, "Not enough space on {tmpdir} to migrate to lmdb " +
                                  "(In {dbmapdir}, " +
                                  "{format_size(required_dbsize + total_entrysize)} is "
                                  "needed but only {format_size(free)} is available)")
    # Lets use dbmap_size if possible, otherwise use required_dbsize
    if free < dbmap_size + total_entrysize:
        dbmap_size = required_dbsize

    progress = 0
    encrypt = False       # Should maybe be a args param
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if be['dbsize'] == 0:
            continue
        log.info(f"Backends exportation {progress*100/total_dbsize:2f}% ({bename})")
        log.debug(f"inst.db2ldif({bename}, None, None, {encrypt}, True, {be['ldifname']})")
        inst.db2ldif(bename, None, None, encrypt, True, be['ldifname'], False)
        if not os.path.isfile(be['ldifname']):
            raise RuntimeError(f"Failed to export backend {bename} into {be['ldifname']}.")

        be['cl5'] = export_changelog(be, 'bdb')
        progress += be['dbsize']
    log.info("Backends exportation 100%")

    log.info("Updating dse.ldif file")
    # switch nsslapd-backend-implement in the dse.ldif
    cfgbe = backends['config']
    dn = cfgbe['dn']
    dse.replace(dn, 'nsslapd-backend-implement', 'mdb')

    # Add the lmdb config entry
    dn = f'cn=mdb,{dn}'
    try:
        dse.delete_dn(dn)
    except Exception:
        pass
    dse.add_entry([
        f"dn: {dn}\n",
        "objectClass: extensibleobject\n",
        "objectClass: top\n",
        "cn: mdb\n",
        f"nsslapd-mdb-max-size: {dbmap_size}\n",
        "nsslapd-mdb-max-readers: 0\n",
        f"nsslapd-mdb-max-dbs: {nbdbis}\n",
        "nsslapd-db-durable-transaction: on\n",
        "nsslapd-search-bypass-filter-test: on\n",
        "nsslapd-serial-lock: on\n"
    ])

    # Reimport all exported backends and changelog
    progress = 0
    encrypt = False       # Should maybe be a args param
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if be['dbsize'] == 0:
            continue
        log.info(f"Backends importation {progress*100/total_dbsize:2f}% ({bename})")
        os.chown(be['ldifname'], uid, gid)
        log.debug(f"inst.ldif2db({bename}, None, None, {encrypt}, {be['ldifname']})")
        inst.ldif2db(bename, None, None, encrypt, be['ldifname'])
        if be['cl5'] is True:
            import_changelog(be, 'mdb')
        progress += be['dbsize']
    dbhome=backends["config"]["dbdir"]
    set_owner(glob.glob(f'{dbhome}/*.mdb'), uid, gid)
    log.info("Backends importation 100%")
    inst.start(post_open=False)
    log.info("Migration from Berkeley database to lmdb is done.")


def dblib_mdb2bdb(inst, log, args):
    global _log
    _log = log
    if args.tmpdir is None:
        tmpdir = get_ldif_dir(inst)
    else:
        tmpdir = args.tmpdir
    try:
        os.makedirs(tmpdir, 0o750, True)
    except OSError as e:
        log.error(f"Failed trying to create the directory {tmpdir} needed to store the ldif files, error: {str(e)}")
        return

    status = get_bdb_impl_status()
    if status is BDB_IMPL_STATUS.UNKNOWN:
        log.warning('Unable to determine if Berkeley Database library is available. If it is not the case, the migration will fail.')
    elif status is BDB_IMPL_STATUS.READ_ONLY:
        log.error('It is not possible to migrate to read-only Berkeley Database library.')
        raise RuntimeError('Berkeley Database library is read-only')
    elif status is BDB_IMPL_STATUS.NONE:
        log.error('Berkeley Database library is not available. Maybe 389-ds-base-bdb rpm should be installed.')
        raise RuntimeError('Berkeley Database library is not available')

    # Cannot use Backends(inst).list() because it requires a connection.
    # lets use directlt the dse.ldif after having stopped the instance

    inst.stop()
    dse = DSEldif(inst)
    uid, gid = get_uid_gid_from_dse(dse)
    backends,dbis = get_backends(log, dse, tmpdir)
    dbmapdir = backends['config']['dbdir']
    dbhome = inst.ds_paths.db_home_dir
    dblib = backends['config']['dblib']

    if dblib == "bdb":
        log.error(f"Instance {inst.serverid} is already configured with bdb.")
        return

    # Remove ldif files and bdb files
    dblib_cleanup(inst, log, args)

    for be in dbis:
        if be is None:
            continue
        id2entry = dbis[be]['id2entry.db']
        if int(id2entry['nbentries']) > 0:
            backends[be]['has_id2entry'] = True

    # Compute the needed space and the lmdb map configuration
    dbmap_size = os.path.getsize(f'{dbmapdir}/{MDB_MAP}')
    # Clearly over evaluated (but better than nothing )
    total_entrysize = dbmap_size

    log.info(f"Required space for LDIF files is about {format_size(total_entrysize)}")
    log.info(f"Required space for bdb files is about {format_size(dbmap_size)}")

    if os.stat(dbmapdir).st_dev != os.stat(tmpdir).st_dev:
        # Ldif and db are on different filesystems
        # Let check that we have enough space for ldif files
        total, used, free = shutil.disk_usage(tmpdir)
        if free < total_entrysize:
            raise OSError(ENOSPC, "Not enough space on {tmpdir} to migrate to bdb " +
                                  "(In {tmpdir}, {format_size(total_entrysize)} bytes "+
                                  "are needed but only {format_size(free)} are available)")
        total_entrysize = 0    # do not count total_entrysize when checking dbmapdir size

    # Let check that we have enough space for the db and ldif files
    total, used, free = shutil.disk_usage(dbmapdir)
    if free < dbmap_size + total_entrysize:
            raise OSError(ENOSPC, "Not enough space on {tmpdir} to migrate to bdb " +
                                  "(In {dbmapdir}, {format_size(dbmap_size+total_entrysize)} "+
                                  "is needed but only {format_size(free)} is available)")

    progress = 0
    encrypt = False       # Should maybe be a args param
    total_dbsize = 0
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if 'has_id2entry' not in be:
            continue
        total_dbsize += 1
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if 'has_id2entry' not in be:
            continue
        log.info(f"Backends exportation {progress*100/total_dbsize:2f}% ({bename})")
        log.debug(f"inst.db2ldif({bename}, None, None, {encrypt}, True, {be['ldifname']})")
        inst.db2ldif(bename, None, None, encrypt, True, be['ldifname'], False)
        be['cl5'] = export_changelog(be, 'mdb')
        progress += 1
    log.info("Backends exportation 100%")
    set_owner(glob.glob(f'{dbmapdir}/*'), uid, gid)

    log.info("Updating dse.ldif file")
    # switch nsslapd-backend-implement in the dse.ldif
    cfgbe = backends['config']
    dn = cfgbe['dn']
    dse.replace(dn, 'nsslapd-backend-implement', 'bdb')

    # bdb entries should still be here

    # Reimport all exported backends and changelog
    progress = 0
    encrypt = False      # Should maybe be a args param
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if 'has_id2entry' not in be:
            continue
        log.info(f"Backends importation {progress*100/total_dbsize:2f}% ({bename})")
        log.debug(f"inst.ldif2db({be['ecbename']}, None, None, {encrypt}, {be['ldifname']})")
        log.debug(f'dbdir={be["dbdir"]}')
        os.chown(be['ldifname'], uid, gid)
        inst.ldif2db(be['ecbename'], None, None, encrypt, be['ldifname'])
        if be['cl5'] is True:
            import_changelog(be, 'bdb')
        set_owner(glob.glob(f'{be["ecdbdir"]}/*'), uid, gid)
        progress += be['dbsize']

    set_owner(glob.glob(f'{dbhome}/__db.*'), uid, gid)
    set_owner(glob.glob(f'{dbmapdir}/__db.*'), uid, gid)
    set_owner(glob.glob(f'{dbhome}/log.*'), uid, gid)
    set_owner(glob.glob(f'{dbmapdir}/log.*'), uid, gid)
    set_owner((f'{dbhome}/DBVERSION', f'{dbmapdir}/DBVERSION', f'{dbhome}/guardian', '{dbmapdir}/guardian'), uid, gid)

    log.info("Backends importation 100%")
    inst.start(post_open=False)
    log.info("Migration from ldbm to Berkeley database is done.")


def rm(path):
    if path is not None:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass


def dblib_cleanup(inst, log, args):
    global _log
    _log = log
    tmpdir = get_ldif_dir(inst)
    dse = DSEldif(inst)
    backends,dbis = get_backends(log, dse, tmpdir)
    dbmapdir = backends['config']['dbdir']
    dbhome = inst.ds_paths.db_home_dir
    dblib = backends['config']['dblib']
    log.info(f"cleanup dbmapdir={dbmapdir} dbhome={dbhome} dblib={dblib}")

    # Remove all ldif and changelog file
    for bename, be in backends.items():
        # Keep only backend associated with a db
        if 'has_id2entry' not in be and be['dbsize'] == 0:
            continue
        # rm(be['ldifname'])
        # rm(be['cl5name'])

    if dblib == "mdb":
        # Looks for bdb database instance subdirectories
        for id2entry in glob.glob(f'{dbmapdir}/*/id2entry.db*'):
            dbdir = Path(id2entry).parent.absolute()
            log.info(f"cleanup removing {dbdir}")
            for f in glob.iglob(f'{dbdir}/*.db*'):
                rm(f)
            for f in [ 'guardian', 'DBVERSION' ]:
                if os.path.isfile(f'{dbdir}/{f}'):
                    rm(f'{dbdir}/{f}')
            os.rmdir(dbdir)

    if dblib == "bdb":
        rm(f'{dbmapdir}/INFO.mdb')
        rm(f'{dbmapdir}/data.mdb')
        rm(f'{dbmapdir}/lock.mdb')
    else:
        for f in glob.iglob(f'{dbhome}/__db.*'):
            rm(f)
        for f in glob.iglob(f'{dbmapdir}/__db.*'):
            rm(f)
        for f in glob.iglob(f'{dbhome}/log.*'):
            rm(f)
        for f in glob.iglob(f'{dbmapdir}/log.*'):
            rm(f)
        rm(f'{dbhome}/DBVERSION')
        rm(f'{dbmapdir}/DBVERSION')
        rm(f'{dbhome}/guardian')
        rm(f'{dbmapdir}/guardian')


def create_parser(subparsers):
    dblib_parser = subparsers.add_parser('dblib', help="database library (i.e bdb/lmdb) migration", formatter_class=CustomHelpFormatter)
    subcommands = dblib_parser.add_subparsers(help="action")

    dblib_bdb2mdb_parser = subcommands.add_parser('bdb2mdb', help='Migrate bdb databases to lmdb', formatter_class=CustomHelpFormatter)
    dblib_bdb2mdb_parser.set_defaults(func=dblib_bdb2mdb)
    dblib_bdb2mdb_parser.add_argument('--tmpdir', help="ldif migration files directory path.")

    dblib_mdb2bdb_parser = subcommands.add_parser('mdb2bdb', help='Migrate lmdb databases to bdb', formatter_class=CustomHelpFormatter)
    dblib_mdb2bdb_parser.set_defaults(func=dblib_mdb2bdb)
    dblib_mdb2bdb_parser.add_argument('--tmpdir', help="ldif migration files directory path.")

    dblib_cleanup_parser = subcommands.add_parser('cleanup', help='Remove migration ldif file and old database', formatter_class=CustomHelpFormatter)
    dblib_cleanup_parser.set_defaults(func=dblib_cleanup)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.dbgen import (
    dbgen_users,
    dbgen_groups,
    dbgen_cos_def,
    dbgen_cos_template,
    dbgen_role,
    dbgen_mod_load,
    dbgen_nested_ldif,
)
from lib389.utils import is_a_dn
from lib389.cli_base import CustomHelpFormatter

DEFAULT_LDIF = "/ldifgen.ldif"

ignore_args = [
    "ldif_file",
    "func",
    "verbose",
    "list",
    "instance",
    "json",
    "remove_all",
]


def get_ldif_dir(instance):
    """
    Get the server's LDIF directory.
    """
    return instance.get_ldif_dir()


def adjust_ldif_name(instance, ldif_name):
    """
    If just a name is provided append it to the server's LDIF directory
    """
    if ldif_name[0] == '.' or ldif_name[0] == '/':
        # Name appears to already be an absolute path
        return ldif_name
    else:
        # Its just a name, add the server's ldif directory
        return instance.get_ldif_dir() + "/" + ldif_name


def display_args(log, args):
    # Display all the options that are being used to generate the ldif file
    log.info(f"\nGenerating LDIF with the following options:")
    for k, v in vars(args).items():
        if k in ignore_args or v is None:
            continue
        k = k.replace("_", "-")  # Restore arg's original name
        log.info(f" - {k}={v}")
    log.info(f" - ldif-file={args.ldif_file}")
    log.info("\nWriting LDIF ...")


def validate_ldif_file(ldif_file, log=None):
    """
    Check if the LDIF file exists.  If interactive then return some error
    text to the caller, otherwise raise an error.
    """
    try:
        f = open(ldif_file, 'w')
        f.close()
        return True
    except PermissionError:
        # File might already exist
        msg = f"The LDIF file ({ldif_file}) exists and can not be overwritten.  Please choose a different file name."
        if log is not None:
            log.info(msg)
        else:
            raise ValueError(msg)
    except FileNotFoundError:
        msg = f"The LDIF file ({ldif_file}) location does not exist.  Please choose a different location."
        if log is not None:
            log.info(msg)
        else:
            raise ValueError(msg)
    except Exception as e:
        msg = f"The LDIF file ({ldif_file}) can not be written: {str(e)}"
        if log is not None:
            log.info(msg)
        else:
            raise ValueError(msg)

    return False


def get_ldif_file_input(log, default_name=DEFAULT_LDIF):
    valid = False
    while not valid:
        file_name = get_input(log, "Enter the new LDIF file name", default_name)
        valid = validate_ldif_file(file_name, log=log)
    return file_name


def get_input(log, msg, default, type="", options=None):
    # Interactive prompt
    display_default = default
    if isinstance(default, bool):
        if default:
            display_default = "yes"
        else:
            display_default = "no"
    while 1:
        if display_default != "":
            val = input(f'\n{msg} [{display_default}]: ')
        else:
            val = input(f'\n{msg}: ')
        if val != '':
            if type == "dn":
                if is_a_dn(val, allow_anon=False):
                    return val
                else:
                    log.info(f"\n --->  The value you entered \"{val}\" is not a valid DN")
                    continue
            elif type == "int":
                if val.isdigit():
                    if int(val) < 1:
                        log.info("\n --->  You must enter number greater than 0")
                        continue
                    return int(val)
                else:
                    log.info(f"\n --->  The number you entered \"{val}\" is not a number")
                    continue
            elif type == "bool":
                if val.lower() == "y" or val.lower() == "yes":
                    return True
                elif val.lower() == "n" or val.lower() == "no":
                    return False
                else:
                    log.info(f"\n --->  Invalid value ({val}), please enter \"yes\" or \"no\".")
                    continue
            else:
                # Just a string, nothing to validate
                if options is not None:
                    if val.lower() not in options:
                        opt_str = ', '.join(options)
                        log.info(f"\n --->  Invalid value ({val}), please enter one of the following types: {opt_str}")
                        continue
                return val
        else:
            # User selected the default value
            return default


def dbgen_create_users(inst, log, args):
    """
    Create a LDIF of user entries
    """
    if args.number is None or args.suffix is None:
        """
        Interactively get all the info ...
        """
        log.info("Missing required parameters '--number' and/or '--suffix', switching to Interactive mode ...")

        # Get the suffix
        args.suffix = get_input(log, "Enter the suffix", "dc=example,dc=com", "dn")

        # Get the parent
        args.parent = get_input(log, "Enter the parent entry for the users", "ou=people,dc=example,dc=com", "dn")

        # Get the number of users to create
        args.number = get_input(log, "Enter the number of users to create", 100000, "int")

        # Confirm the RDN attribute
        args.rdn_cn = get_input(log, "Do you want to use \"cn\" instead of \"uid\" for the entry RDN attribute (yes/no)", False, "bool")

        # Create generic entries
        args.generic = get_input(log, "Create generic entries that can be used with \"ldclt\" (yes/no)", False, "bool")

        # get offset size
        if args.generic:
            args.start_idx = get_input(log, "Choose the starting index for the generic user entries", 0, "int")

        # localize the data
        args.localize = get_input(log, "Do you want to localize the LDIF data (yes/no)", False, "bool")

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        args.ldif_file = adjust_ldif_name(inst, args.ldif_file)
        validate_ldif_file(args.ldif_file)

    display_args(log, args)
    dbgen_users(inst, args.number, args.ldif_file, args.suffix, generic=args.generic, parent=args.parent, startIdx=args.start_idx, rdnCN=False, pseudol10n=args.localize)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_groups(inst, log, args):
    """
    Create static groups and their members
    """

    if args.number is None or args.suffix is None:
        """
        Interactively get all the info ...
        """
        log.info("Missing required parameters '--number' and/or '--suffix', switching to Interactive mode ...")

        # Get the number of users to create
        args.number = get_input(log, "Enter the number of groups to create", 1, "int")

        # Get the suffix
        args.suffix = get_input(log, "Enter the suffix", "dc=example,dc=com", "dn")

        # Get the parent
        args.parent = get_input(log, "Enter the parent entry to add the groups under", args.suffix, "dn")

        # Get the membership attr
        args.member_attr = get_input(log, "Enter the attribute to use for the group membership", "uniquemember")

        # Number of members
        args.num_members = get_input(log, "Enter the number of members to add to the group", 10000, "int")

        # Create member entries
        args.create_members = get_input(log, "Do you want to create the member entries (yes/no)", True, "bool")

        # member entries parent
        if args.create_members:
            args.member_parent = get_input(log, "Enter the parent entry to add the users under", args.suffix, "dn")

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        validate_ldif_file(args.ldif_file)

    props = {
        "name": args.NAME,
        "parent": args.parent,
        "suffix": args.suffix,
        "number": args.number,
        "numMembers": args.num_members,
        "createMembers": args.create_members,
        "memberParent": args.member_parent,
        "membershipAttr": args.member_attr,
    }

    display_args(log, args)
    dbgen_groups(inst, args.ldif_file, props)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_cos_def(inst, log, args):
    """
    Create a COS definition
    """
    if args.type is None or args.parent is None or len(args.cos_attr) == 0 or \
        ((args.type == "classic" or args.type == "indirect") and args.cos_specifier is None) \
        or ((args.type == "classic" or args.type == "pointer") and args.cos_template is None):
        log.info("Missing some required parameters '--parent',  '--type', '--cos-specifier', or '--cos-template', switching to Interactive mode ...")

        # Get the number of users to create
        args.type = get_input(log, "Type of COS definition: \"classic\", \"pointer\", or \"indirect\"",
                              "classic",  options=["classic", "pointer", "indirect"])

        # Get the parent
        args.parent = get_input(log, "Enter the parent entry to add the COS definition under", "dc=example,dc=com", "dn")

        # Create parent
        args.create_members = get_input(log, "Do you want to create the parent entry (yes/no)", True, "bool")

        # COS specifier
        if args.type == "classic" or args.type == "indirect":
            args.cos_specifier = get_input(log, "Enter the COS specifier attribute", "description")

        # COS template DN
        if args.type == "classic" or args.type == "pointer":
            args.cos_template = get_input(log, "Enter the COS Template DN", "cn=COS Template Entry,dc=example,dc=com", "dn")

        # Gather the COS attributes
        while True:
            val = get_input(log, "Enter COS attributes, press Enter when finished", "")
            if val == "" and len(args.cos_attr) > 0:
                break
            args.cos_attr.append(val)

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        validate_ldif_file(args.ldif_file)

    props = {
        "cosType": args.type,
        "defName": args.NAME,
        "defParent": args.parent,
        "defCreateParent": args.create_parent,
        "cosSpecifier": args.cos_specifier,
        "cosAttrs": args.cos_attr,
        "tmpName": args.cos_template
    }

    display_args(log, args)
    dbgen_cos_def(inst, args.ldif_file, props)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_cos_tmp(inst, log, args):
    """
    Create a COS template entry
    """
    if args.parent is None or args.cos_priority is None or args.cos_attr_val is None:
        log.info("Missing required parameters '--parent', '--cos-priority' or '--cos-attr-val', switching to Interactive mode ...")

        # Get the parent
        args.parent = get_input(log, "Enter the parent entry to add the COS template under", "dc=example,dc=com", "dn")

        # Create parent
        args.create_parent = get_input(log, "Do you want to create the parent entry (yes/no)", True, "bool")

        # Get the COS priority
        args.cos_priority = get_input(log, "Enter the COS priority for this template", "0", "int")

        # Get the attribute value pair
        args.cos_attr_val = get_input(log, "Enter the attribute and value pair.  Use this format: \"ATTRIBUTE:VALUE\"", "postalcode:19605")

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        validate_ldif_file(args.ldif_file)

    props = {
        "tmpName": args.NAME,
        "tmpParent": args.parent,
        "tmpCreateParent": args.create_parent,
        "cosPriority": args.cos_priority,
        "cosTmpAttrVal": args.cos_attr_val,
    }

    display_args(log, args)
    dbgen_cos_template(inst,  args.ldif_file, props)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_role(inst, log, args):
    """
    Create a Role
    """
    if args.type is None or args.parent is None or \
        (args.type == "filtered" and args.filter is None) or \
        (args.type == "nested" and len(args.role_dn) == 0):
        log.info("Missing some required parameters '--type', '--parent'. '--filter', '--role-dn', switching to Interactive mode ...")

        # Get the number of users to create
        args.type = get_input(log, "Type of Role: \"managed\", \"filtered\", or \"nested\"",
                              "managed",  options=["managed", "filtered", "nested"])

        # Get the parent
        args.parent = get_input(log, "Enter the parent entry to add the Role under", "dc=example,dc=com", "dn")

        # Create parent
        args.create_parent = get_input(log, "Do you want to create the parent entry (yes/no)", True, "bool")

        # Role filter
        if args.type == "filtered":
            args.filter = get_input(log, "Enter the Role filter", "cn=some_value")

        # Role DN (nested only)
        if args.type == "nested":
            while True:
                val = get_input(log, "Enter the Role DN", "cn=some other role,dc=example,dc=com", "dn")
                if val == "" and len(args.role_dn) > 0:
                    break
                args.role_dn.append(val)

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        validate_ldif_file(args.ldif_file)

    props = {
        "role_type": args.type,
        "role_name": args.NAME,
        "parent": args.parent,
        "createParent": args.create_parent,
        "filter": args.filter,
        "role_list": args.role_dn
    }

    display_args(log, args)
    dbgen_role(inst, args.ldif_file, props)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_mods(inst, log, args):
    """
    Create a LDIF file of update operations that can be consumed by ldapmodify

    There are a lot of options here for creating different types of modification
    LDIFs.  One technique/option is that you can work with existing users.  Use
    dbgen to generate a large generic user ldif, import it, then you can create
    a modification LDIF that will work on that database.  It's a lot faster than
    using ldapmodify to add 1 million first then modify those entries.

    The other nice option is that you can randomize the operations, but this does
    introduce a potential for some of the operations to fail.  You could delete
    an entry before you modify it, etc...
    """

    if args.num_users is None or args.parent is None:
        log.info("Missing required parameters '--num-users' and/or '--parent', switching to Interactive mode ...")

        # Create users
        args.create_users = get_input(log, "Do you want to create the user entries (yes/no)", True, "bool")

        # Delete users
        args.delete_users = get_input(log, "Do you want to delete all the user entries at the end (yes/no)", True, "bool")

        # Get the number of entries
        args.num_users = get_input(log, "Enter the number of user entries that can be modified", "100000", "int")

        # Get the parent entry
        args.parent = get_input(log, "The DN of the parent entry where the user entries are located", "ou=people,dc=example,dc=com")

        # Create parent
        args.create_parent = get_input(log, "Create the parent entry (yes/no)", True, "bool")

        # Add users
        args.add_users = get_input(log, "The number of users to add during the load", 0, "int")

        # Delete users
        args.del_users = get_input(log, "The number of users to delete during the load", 0, "int")

        # modrdn users
        args.modrdn_users = get_input(log, "The number of users to modrdn during the load", 0, "int")

        # modify users
        args.mod_users = get_input(log, "The number of users to modify during the load", 0, "int")

        # Modification attributes
        args.mod_attrs = get_input(log, "List of attributes that will be randomly chosen from when modifying an entry", "description cn")
        args.mod_attrs = args.mod_attrs.split(' ')

        # Randomize the load
        args.randomize = get_input(log, "Randomly perform the specified add, mod, delete, and modrdn operations (yes/no)", True, "bool")

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        validate_ldif_file(args.ldif_file)

    props = {
        "createUsers": args.create_users,
        "deleteUsers": args.delete_users,
        "numUsers": int(args.num_users),
        "parent": args.parent,
        "createParent": args.create_parent,
        "addUsers": int(args.add_users),
        "delUsers": int(args.del_users),
        "modrdnUsers": int(args.modrdn_users),
        "modUsers": int(args.mod_users),
        "random": args.randomize,
        "modAttrs": args.mod_attrs
    }

    display_args(log, args)
    dbgen_mod_load(args.ldif_file, props)
    log.info(f"Successfully created LDIF file: {args.ldif_file}")


def dbgen_create_nested(inst, log, args):
    """
    Create a cascading/fractal tree.  Every node splits in half and keeps
    branching out in that fashion until all the entries are used up.
    """

    if  args.num_users is None or args.node_limit is None or args.suffix is None:
        # Num users
        args.num_users = get_input(log, "The number of users to add during the load", 1000000, "int")

        # Node limit
        args.node_limit = get_input(log, "The total number of user entries to create under each node/subtree", 500, "int")

        # Suffix
        args.suffix = get_input(log, "Enter the suffix", "dc=example,dc=com", "dn")

        # Get the output LDIF file name
        args.ldif_file = get_ldif_file_input(log, default_name=get_ldif_dir(inst) + DEFAULT_LDIF)
    else:
        args.ldif_file = adjust_ldif_name(inst, args.ldif_file)
        validate_ldif_file(args.ldif_file)

    props = {
        "numUsers": int(args.num_users),
        "nodeLimit": int(args.node_limit),
        "suffix": args.suffix,
    }

    display_args(log, args)
    node_count = dbgen_nested_ldif(inst, args.ldif_file, props)
    log.info(f"Successfully created nested LDIF file ({args.ldif_file}) containing {node_count} nodes/subtrees")


def create_parser(subparsers):
    db_gen_parser = subparsers.add_parser('ldifgen', help="LDIF generator to make sample LDIF files for testing", formatter_class=CustomHelpFormatter)
    subcommands = db_gen_parser.add_subparsers(help="action")

    # Create just users
    dbgen_users_parser = subcommands.add_parser('users', help='Generate a LDIF containing user entries', formatter_class=CustomHelpFormatter)
    dbgen_users_parser.set_defaults(func=dbgen_create_users)
    dbgen_users_parser.add_argument('--number', help="The number of users to create.")
    dbgen_users_parser.add_argument('--suffix', help="The database suffix where the entries will be created.")
    dbgen_users_parser.add_argument('--parent', help="The parent entry that the user entries should be created under.  If not specified, the entries are stored under random Organizational Units.")
    dbgen_users_parser.add_argument('--generic', action='store_true', help="Create generic entries in the format of \"uid=user####\".  These entries are also compatible with ldclt.")
    dbgen_users_parser.add_argument('--start-idx', default=0, help="For generic LDIF's you can choose the starting index for the user entries.  The default is \"0\".")
    dbgen_users_parser.add_argument('--rdn-cn', action='store_true', help="Use the attribute \"cn\" as the RDN attribute in the DN instead of \"uid\"")
    dbgen_users_parser.add_argument('--localize', action='store_true', help="Localize the LDIF data")
    dbgen_users_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create static groups
    dbgen_groups_parser = subcommands.add_parser('groups', help='Generate a LDIF containing groups and members', formatter_class=CustomHelpFormatter)
    dbgen_groups_parser.set_defaults(func=dbgen_create_groups)
    dbgen_groups_parser.add_argument('NAME', help="The group name.")
    dbgen_groups_parser.add_argument('--number', default=1, help="The number of groups to create.")
    dbgen_groups_parser.add_argument('--suffix', help="The database suffix where the groups will be created.")
    dbgen_groups_parser.add_argument('--parent', help="The parent entry that the group entries should be created under.  If not specified the groups are stored under the suffix.")
    dbgen_groups_parser.add_argument('--num-members', default="10000", help="The number of members in the group.  Default is 10000")
    dbgen_groups_parser.add_argument('--create-members', action='store_true', help="Create the member user entries.")
    dbgen_groups_parser.add_argument('--member-parent', help="The entry DN that the members should be created under.  The default is the suffix entry.")
    dbgen_groups_parser.add_argument('--member-attr', default="uniquemember", help="The membership attribute to use in the group.  Default is \"uniquemember\".")
    dbgen_groups_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create a COS definition
    dbgen_cos_def_parser = subcommands.add_parser('cos-def', help='Generate a LDIF containing a COS definition (classic, pointer, or indirect)', formatter_class=CustomHelpFormatter)
    dbgen_cos_def_parser.set_defaults(func=dbgen_create_cos_def)
    dbgen_cos_def_parser.add_argument('NAME', help="The COS definition name.")
    dbgen_cos_def_parser.add_argument('--type', help="The COS definition type: \"classic\", \"pointer\", or \"indirect\".")
    dbgen_cos_def_parser.add_argument('--parent', help="The parent entry that the COS definition should be created under.")
    dbgen_cos_def_parser.add_argument('--create-parent', action='store_true', help="Create the parent entry")
    dbgen_cos_def_parser.add_argument('--cos-specifier', help="Used in a classic COS definition, this attribute located in the user entry is used to select which COS template to use.")
    dbgen_cos_def_parser.add_argument('--cos-template', help="The DN of the COS template entry, only used for \"classic\" and \"pointer\" COS definitions.")
    dbgen_cos_def_parser.add_argument('--cos-attr', nargs='*', default=[], help="A list of attributes which defines which attribute the COS generates values for.")
    dbgen_cos_def_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create a COS Template
    dbgen_cos_tmp_parser = subcommands.add_parser('cos-template', help='Generate a LDIF containing a COS template', formatter_class=CustomHelpFormatter)
    dbgen_cos_tmp_parser.set_defaults(func=dbgen_create_cos_tmp)
    dbgen_cos_tmp_parser.add_argument('NAME', help="The COS template name.")
    dbgen_cos_tmp_parser.add_argument('--parent', help="The DN of the entry to store the COS template entry under.")
    dbgen_cos_tmp_parser.add_argument('--create-parent', action='store_true', help="Create the parent entry")
    dbgen_cos_tmp_parser.add_argument('--cos-priority', type=int, help="Sets the priority of this conflicting/competing COS templates.")
    dbgen_cos_tmp_parser.add_argument('--cos-attr-val', help="defines the attribute and value that the template provides.")
    dbgen_cos_tmp_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create Role entries
    dbgen_roles_parser = subcommands.add_parser('roles', help='Generate a LDIF containing a role entry (managed, filtered, or indirect)', formatter_class=CustomHelpFormatter)
    dbgen_roles_parser.set_defaults(func=dbgen_create_role)
    dbgen_roles_parser.add_argument('NAME', help="The Role name.")
    dbgen_roles_parser.add_argument('--type', help="The Role type: \"managed\", \"filtered\", or \"nested\".")
    dbgen_roles_parser.add_argument('--parent', help="The DN of the entry to store the Role entry under")
    dbgen_roles_parser.add_argument('--create-parent', action='store_true', help="Create the parent entry")
    dbgen_roles_parser.add_argument('--filter', help="A search filter for gathering Role members.  Required for a \"filtered\" role.")
    dbgen_roles_parser.add_argument('--role-dn', nargs='*', default=[], help="A DN of a role entry that should be included in this role.  Used for \"nested\" roles only.")
    dbgen_roles_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create a modification LDIF
    dbgen_mod_load_parser = subcommands.add_parser('mod-load', help='Generate a LDIF containing modify operations.  This is intended to be consumed by ldapmodify.', formatter_class=CustomHelpFormatter)
    dbgen_mod_load_parser.set_defaults(func=dbgen_create_mods)
    dbgen_mod_load_parser.add_argument('--create-users', action='store_true', help="Create the entries that will be modified or deleted.  By default the script assumes the user entries already exist.")
    dbgen_mod_load_parser.add_argument('--delete-users', action='store_true', help="Delete all the user entries at the end of the LDIF.")
    dbgen_mod_load_parser.add_argument('--num-users', type=int, help="The number of user entries that will be modified or deleted")
    dbgen_mod_load_parser.add_argument('--parent', help="The DN of the parent entry where the user entries are located.")
    dbgen_mod_load_parser.add_argument('--create-parent', action='store_true', help="Create the parent entry")
    dbgen_mod_load_parser.add_argument('--add-users', default=100, help="The number of additional entries to add during the load.")
    dbgen_mod_load_parser.add_argument('--del-users', default=100, help="The number of entries to delete during the load.")
    dbgen_mod_load_parser.add_argument('--modrdn-users', default=100, help="The number of entries to perform a modrdn operation on.")
    dbgen_mod_load_parser.add_argument('--mod-users', default=100, help="The number of entries to modify.")
    dbgen_mod_load_parser.add_argument('--mod-attrs', nargs="*", default=['description'], help="List of attributes the script will randomly choose from when modifying an entry.  The default is \"description\".")
    dbgen_mod_load_parser.add_argument('--randomize', action='store_true', help="Randomly perform the specified add, mod, delete, and modrdn operations")
    dbgen_mod_load_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")

    # Create a heavily nested LDIF
    dbgen_nested_parser = subcommands.add_parser('nested', help='Generate a heavily nested database LDIF in a cascading/fractal tree design', formatter_class=CustomHelpFormatter)
    dbgen_nested_parser.set_defaults(func=dbgen_create_nested)
    dbgen_nested_parser.add_argument('--num-users', help="The total number of user entries to create in the entire LDIF (does not include the container entries).")
    dbgen_nested_parser.add_argument('--node-limit', help="The total number of user entries to create under each node/subtree")
    dbgen_nested_parser.add_argument('--suffix', help="The suffix DN for the LDIF")
    dbgen_nested_parser.add_argument('--ldif-file', default="ldifgen.ldif", help=f"The LDIF file name.  Default location is the server's LDIF directory using the name 'ldifgen.ldif'")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap

from lib389._mapped_object import DSLdapObjects, DSLdapObject


class Referral(DSLdapObject):
    def __init__(self, instance, dn=None):
        super(Referral, self).__init__(instance, dn)
        self._rdn_attribute = "cn"
        self._must_attributes = ["ref"]
        self._create_objectclasses = ['referral', 'nsContainer']
        self._protected = False
        managedsait_ctrl = ldap.controls.simple.ManageDSAITControl()
        self._server_controls = [managedsait_ctrl]
        self._client_controls = None

class Referrals(DSLdapObjects):
    def __init__(self, instance, basedn):
        super(Referrals, self).__init__(instance)
        self._objectclasses = ['referral']
        self._filterattrs = 'cn'
        self._childobject = Referral
        self._basedn = basedn
        managedsait_ctrl = ldap.controls.simple.ManageDSAITControl()
        self._server_controls = [managedsait_ctrl]
        self._client_controls = None




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.plugins import AttributeUniquenessPlugin, AttributeUniquenessPlugins
from lib389.cli_conf import (generic_object_edit, generic_object_add)
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'enabled': 'nsslapd-pluginenabled',
    'attr_name': 'uniqueness-attribute-name',
    'subtree': 'uniqueness-subtrees',
    'exclude_subtree': 'uniqueness-exclude-subtrees',
    'across_all_subtrees': 'uniqueness-across-all-subtrees',
    'top_entry_oc': 'uniqueness-top-entry-oc',
    'subtree_entries_oc': 'uniqueness-subtree-entries-oc'
}

PLUGIN_DN = "cn=plugins,cn=config"


def attruniq_list(inst, basedn, log, args):
    log = log.getChild('attruniq_list')
    plugins = AttributeUniquenessPlugins(inst)
    result = []
    result_json = []
    for plugin in plugins.list():
        if args.json:
            result_json.append(json.loads(plugin.get_all_attrs_json()))
        else:
            result.append(plugin.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json},  indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Attribute Uniqueness plugin instances")


def attruniq_add(inst, basedn, log, args):
    log = log.getChild('attruniq_add')
    props = {'cn': args.NAME}
    # We require a subtree, or a target Objectclass
    if args.subtree_entries_oc is None and args.subtree is None:
        raise ValueError("A attribute uniqueness configuration requires a 'subtree' or 'subtree-entries-oc' to be set")

    generic_object_add(AttributeUniquenessPlugin, inst, log, args, arg_to_attr, basedn=PLUGIN_DN, props=props)


def attruniq_edit(inst, basedn, log, args):
    log = log.getChild('attruniq_edit')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)
    generic_object_edit(plugin, log, args, arg_to_attr)


def attruniq_show(inst, basedn, log, args):
    log = log.getChild('attruniq_show')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)

    if not plugin.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.name)
    if args and args.json:
        log.info(plugin.get_all_attrs_json())
    else:
        log.info(plugin.display())


def attruniq_del(inst, basedn, log, args):
    log = log.getChild('attruniq_del')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)
    plugin.delete()
    log.info("Successfully deleted the %s", plugin.dn)


def attruniq_enable(inst, basedn, log, args):
    log = log.getChild('attruniq_enable')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)
    if plugin.status():
        log.info("Plugin '%s' already enabled" % plugin.rdn)
    else:
        plugin.enable()
        log.info("Successfully enabled the %s", plugin.dn)


def attruniq_disable(inst, basedn, log, args):
    log = log.getChild('attruniq_disable')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)
    if not plugin.status():
        log.info("Plugin '%s' already disabled" % plugin.rdn)
    else:
        plugin.disable()
        log.info("Successfully disabled the %s", plugin.dn)


def attruniq_status(inst, basedn, log, args):
    log = log.getChild('attruniq_status')
    plugins = AttributeUniquenessPlugins(inst)
    plugin = plugins.get(args.NAME)
    if plugin.status() is True:
        log.info("Plugin '%s' is enabled" % plugin.rdn)
    else:
        log.info("Plugin '%s' is disabled" % plugin.rdn)


def _add_parser_args(parser):
    parser.add_argument('NAME', help='The name of the plug-in configuration record. (cn) You can use any string, '
                                     'but "attribute_name Attribute Uniqueness" is recommended.')
    parser.add_argument('--enabled', choices=['on', 'off'],
                        help='Identifies whether or not the config is enabled.')
    parser.add_argument('--attr-name', nargs='+',
                        help='Sets the name of the attribute whose values must be unique. '
                             'This attribute is multi-valued. (uniqueness-attribute-name)')
    parser.add_argument('--subtree', nargs='+',
                        help='Sets the DN under which the plug-in checks for uniqueness of '
                             'the attributes value. This attribute is multi-valued (uniqueness-subtrees)')
    parser.add_argument('--exclude-subtree', nargs='+',
                        help='Sets subtrees that should be excluded from attribute uniqueness checks. '
                             'This attribute is multi-valued (uniqueness-exclude-subtrees)')
    parser.add_argument('--across-all-subtrees', choices=['on', 'off'], type=str.lower,
                        help='If enabled (on), the plug-in checks that the attribute is unique across all subtrees '
                             'set. If you set the attribute to off, uniqueness is only enforced within the subtree '
                             'of the updated entry (uniqueness-across-all-subtrees)')
    parser.add_argument('--top-entry-oc',
                        help='Verifies that the value of the attribute set in uniqueness-attribute-name '
                             'is unique in this subtree (uniqueness-top-entry-oc)')
    parser.add_argument('--subtree-entries-oc',
                        help='Verifies if an attribute is unique, if the entry contains the object class '
                             'set in this parameter (uniqueness-subtree-entries-oc)')


def create_parser(subparsers):
    attruniq = subparsers.add_parser('attr-uniq', help='Manage and configure Attribute Uniqueness plugin', formatter_class=CustomHelpFormatter)
    subcommands = attruniq.add_subparsers(help='action')
    # We can't use the add_generic_plugin_parsers as we need named sub instances.

    list = subcommands.add_parser('list', help='Lists available plugin configs', formatter_class=CustomHelpFormatter)
    list.set_defaults(func=attruniq_list)

    add = subcommands.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=attruniq_add)
    _add_parser_args(add)

    edit = subcommands.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=attruniq_edit)
    _add_parser_args(edit)

    show = subcommands.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show.add_argument('NAME', help='The name of the plug-in configuration record')
    show.set_defaults(func=attruniq_show)

    delete = subcommands.add_parser('delete', help='Delete the config entry', formatter_class=CustomHelpFormatter)
    delete.add_argument('NAME', help='The name of the plug-in configuration record')
    delete.set_defaults(func=attruniq_del)

    enable = subcommands.add_parser('enable', help='enable plugin', formatter_class=CustomHelpFormatter)
    enable.add_argument('NAME', help='The name of the plug-in configuration record')
    enable.set_defaults(func=attruniq_enable)

    disable = subcommands.add_parser('disable', help='disable plugin', formatter_class=CustomHelpFormatter)
    disable.add_argument('NAME', help='The name of the plug-in configuration record')
    disable.set_defaults(func=attruniq_disable)

    status = subcommands.add_parser('status', help='display plugin status', formatter_class=CustomHelpFormatter)
    status.add_argument('NAME', help='The name of the plug-in configuration record')
    status.set_defaults(func=attruniq_status)
